#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include "mex.h"

#include <limits.h>
#include <stdio.h>
#include <stdlib.h>

#include "BinaryHeap.h"

/**
 * Allocates memory, initialises and returns a BinaryHeap.
 */
BinaryHeap *restrict
createBinaryHeap(const uint64_t maxNumberOfElements)
{
	BinaryHeap* heap = malloc( sizeof(BinaryHeap) );

	heap->numberOfElements = 0;
	heap->maxNumberOfElements = maxNumberOfElements;
	heap->colToIndex = malloc( sizeof(uint64_t) * maxNumberOfElements );
	heap->elements = malloc( sizeof(HeapElement) * maxNumberOfElements );

	return heap;
}

/**
 * Free()s the memory allocated for the heap.
 */
void
freeBinaryHeap(BinaryHeap *restrict heap)
{
	free(heap->elements);
	free(heap);
}

/**
 * Inserts a column with a given score into the heap.
 * Does not check whether the column already exists;
 * if it already exists, the behaviour of the heap is undefined.
 */
void
insert(BinaryHeap *restrict heap,
	   const uint64_t score,
	   const uint64_t column)
{
	if (heap->numberOfElements >= heap->maxNumberOfElements) {
		//mexPrintf("Trying to insert too many elements into binary heap.\n");
		return;
	}

	heap->elements[heap->numberOfElements].score = score;
	heap->elements[heap->numberOfElements].column = column;
	heap->colToIndex[column] = heap->numberOfElements;
	heap->numberOfElements += 1;

	siftUp(heap, heap->numberOfElements-1);
}

/**
 * Removes the biggest element from the heap.
 */
HeapElement
deleteMax(BinaryHeap *restrict heap)
{
	if (heap->numberOfElements == 0) {
		//mexPrintf("Trying to delete from empty heap.\n");
		return heap->elements[0];
	}

	HeapElement root = heap->elements[0];
	
	heap->elements[0] = heap->elements[heap->numberOfElements - 1];
	heap->colToIndex[heap->elements[0].column] = 0;
	siftDown(heap, 0);
	heap->numberOfElements -= 1;

	return root;
}

uint64_t
peakMax(BinaryHeap *restrict heap)
{
	if (heap->numberOfElements == 0) {
		//mexPrintf("Trying to delete from empty heap.\n");
		return UINT64_MAX;
	}

	HeapElement root = heap->elements[0];
	return root.score;
}

/**
 * Deletes the given column from the heap.
 */
void
deleteColumn(BinaryHeap *restrict heap,
			 const uint64_t column)
{
	updateKey(heap, UINT64_MAX, column);
	deleteMax(heap);
}

/**
 * Internal function to move an element of the heap upwards.
 */
void
siftUp(BinaryHeap *restrict heap,
	   uint64_t elementIndex)
{
	if (elementIndex == 0) {
		return;
	}

	uint64_t parentIndex = (elementIndex-1)/2;

	if (heapElementIsLarger(heap, elementIndex, parentIndex)) {
		HeapElement tmpParent = heap->elements[parentIndex];
		heap->elements[parentIndex] = heap->elements[elementIndex];
		heap->elements[elementIndex] = tmpParent;

		heap->colToIndex[heap->elements[parentIndex].column] = parentIndex;
		heap->colToIndex[heap->elements[elementIndex].column] = elementIndex;

		siftUp(heap, parentIndex);
	}
}

/**
 * Internal function to move an element of the heap downwards.
 */
void
siftDown(BinaryHeap *restrict heap,
		 uint64_t elementIndex)
{
	if (elementIndex > heap->numberOfElements) {
		//mexPrintf("Trying to sift down an element that does not exist.\n");
		return;
	}

	// check whether we have a leaf
	if ( ! (2*(elementIndex+1)-1 < heap->numberOfElements)) {
		return;
	}

	// find smallest child
	uint64_t smallestChild = 2*(elementIndex+1)-1; // first child
	if ( smallestChild+1 < heap->numberOfElements
			&& heapElementIsLarger(heap, smallestChild+1, smallestChild)) {
		// check whether second child exists and is smaller
		smallestChild += 1;
	}

	if (heapElementIsLarger(heap, smallestChild, elementIndex)) {
		HeapElement tmpParent = heap->elements[elementIndex];
		heap->elements[elementIndex] = heap->elements[smallestChild];
		heap->elements[smallestChild] = tmpParent;

		heap->colToIndex[heap->elements[elementIndex].column] = elementIndex;
		heap->colToIndex[heap->elements[smallestChild].column] = smallestChild;

		siftDown(heap, smallestChild);
	}
}

/**
 * Returns true if the score of the element at elementIndex1 is smaller
 * than the on at elementIndex2. If both scores are equal, it returns
 * true if the first element's column number is smaller.
 */
bool
heapElementIsLarger(BinaryHeap *restrict heap,
					const uint64_t elementIndex1,
					const uint64_t elementIndex2)
{
	if (heap->elements[elementIndex1].score != heap->elements[elementIndex2].score) {
		return (heap->elements[elementIndex1].score > heap->elements[elementIndex2].score);
	} else {
		return (heap->elements[elementIndex1].column < heap->elements[elementIndex2].column);
	}
}

/**
 * Updates the score of the given column inside the heap.
 * Notice that this function handles both increaseKey() and
 * decreaseKey() functionalities.
 */
void
updateKey(BinaryHeap *restrict heap,
		  const uint64_t score,
		  const uint64_t column)
{
	uint64_t elementIndex = heap->colToIndex[column];

	if (heap->elements[elementIndex].column != column) {
		//mexPrintf("Cannot update key, element was already deleted.\n");
		return;
	}

	if (heap->elements[elementIndex].score > score) { // decrease key
		heap->elements[elementIndex].score = score;
		siftDown(heap, elementIndex);
	} else { // increase key
		heap->elements[elementIndex].score = score;
		siftUp(heap, elementIndex);
	}
}

uint64_t
getScore(BinaryHeap *restrict heap,
		 const uint64_t key)
{
	const uint64_t elementIndex = heap->colToIndex[key];
	if (heap->elements[elementIndex].column != key) {
		//mexPrintf("Cannot get score, key was already deleted.\n");
		return UINT64_MAX;
	}
	return heap->elements[elementIndex].score;
}

void
increaseKey(BinaryHeap *restrict heap,
			const uint64_t key)
{
	const uint64_t oldScore = getScore(heap, key);
	updateKey(heap, oldScore+1, key);
}

void
decreaseKey(BinaryHeap *restrict heap,
			const uint64_t key)
{
	const uint64_t oldScore = getScore(heap, key);
	if (oldScore > 0) {
		updateKey(heap, oldScore-1, key);
	}
}

