function [ A, l, r ] = ClosestNonnegativeRoundingRankOne(B,varargin)%,k,s,maxIter,epsilon,x_initial,y_initial)
%CLOSESTNONNEGATIVEROUNDINGRANKONE Heuristically computes the closest matrix of B of nonnegative rounding rank 1.
% as for the optional parameters one can pass the following things (in this order):
%	k	-	the steepness of the logistic function
%	s	-	the minimal stepsize for the gradient descent
%	maxIter	-	maximum number of iterations
%	epsilon	-	threshold after which we stop improving our solution
%	x_initial	-	initial solution for gradient descent for x
%	y_initial	-	initial solution for gradient descent for y
%
%   This code is free to use for scientific purposes provided that the
%   original paper is cited.
%
%   Copyright Pauli Miettinen & Stefan Neumann, 2016

m = size(B,1);
n = size(B,2);

if nargin < 2
	k = 8; % steepness of the logistic function
else
	k = varargin{1};
end
if nargin < 3
	s = 0.00001; % stepsize for the gradient descent
else
	s = varargin{2};
end
if nargin < 4
	maxIter = 1000;
else
	maxIter = varargin{3};
end
if nargin < 5
	epsilon = 0.00001; % we will stop once the error decreases by less than epsilon
else
	epsilon = varargin{4};
end
if nargin < 6
	x_initial = sqrt(1/2)*ones(m,1) - epsilon;
else
	x_initial = varargin{5};
end
if nargin < 7
	y_initial = sqrt(1/2)*ones(n,1) - epsilon;
else
	y_initial = varargin{6};
end

iter = 0;
improvement = Inf;

MinValue = 0.00001; % smallest value allowed in x or y

x = zeros(m,1);
y = zeros(n,1);
x_cand = x_initial;
y_cand = y_initial;

oldScore = objectiveFunction(B,x_cand,y_cand,k);
initialScore = oldScore;

while iter < maxIter && improvement > epsilon
	x = x_cand;
	y = y_cand;

	oldScore = objectiveFunction(B,x,y,k);

%	disp('improving x');
	x_cand = improveX(B, x, y, maxIter, epsilon, MinValue, k, s, m, n);
%	disp('improving y');
	y_cand = improveY(B, x, y, maxIter, epsilon, MinValue, k, s, m, n);

	newScore = objectiveFunction(B,x_cand,y_cand,k);

	improvement = oldScore - newScore;
	iter = iter + 1;
end

initialScore
oldScore = objectiveFunction(B,x,y,k)

l = x;
r = y;
A = l * r';
end

function [ f ] = objectiveFunction(B, x, y, k)
f = norm( B - (1 + exp( -k * (x*y' - 1/2) )).^(-1) );
end

function [ xNew, scoreNew ] = improveX(B, x, y, maxIter, epsilon, MinValue, k, s, m, n)
iter = 0;
improvement = Inf;

while iter < maxIter && improvement > epsilon
	oldScore = objectiveFunction(B,x,y,k);
	
	% compute the gradient of f(x)
	delta_x = zeros(m,1);
	for i=1:m
		subvec = exp(-k * (x(i) * y - 1/2));
		vec = (1 + subvec).^(-1);
		
		v1 = 2 * ( B(i,:) - vec' ) .* (vec.^2)';
		v2 = (-k) * subvec .* y;

		delta_x(i) = dot(v1,v2);
%		for j=1:n
%			a1 = exp(-k * (x(i) * y(j) - 1/2));
%			a2 = (1 + a1)^(-1);
%			delta_x(i) = delta_x(i) - 2 * (B(i,j)-a2) * a2^2 * a1 * k * y(j);
%		end
	end

	% update the candidate of x
	step = determineStepSizeX(x, delta_x, s, B, y, k);
	x_cand = x - step*delta_x;
	x_cand = max(x_cand, MinValue); % make sure x does not become negative

	newScore = objectiveFunction(B,x_cand,y,k);

	improvement = oldScore - newScore;
	iter = iter + 1;

	if improvement > epsilon
		x = x_cand;
	end
end

xNew = x;

end

function [ step ] = determineStepSizeX(x, delta_x, s, B, y, k)
	step = s;
	curScore = objectiveFunction(B,x,y,k);
	newScore = 0;

	while curScore > newScore
		curScore = newScore;

		step = 2*step;
		x_cand = x - step*delta_x;

		newScore = objectiveFunction(B,x_cand,y,k);
	end

	step = step/2;
end

function [ step ] = determineStepSizeY(y, delta_y, s, B, x, k)
	step = s;
	curScore = objectiveFunction(B,x,y,k);
	newScore = 0;

	while curScore > newScore
		curScore = newScore;

		step = 2*step;
		y_cand = y - step*delta_y;

		newScore = objectiveFunction(B,x,y_cand,k);
	end

	step = step/2;
end

function [ yNew, newScore ] = improveY(B, x, y, maxIter, epsilon, MinValue, k, s, m, n)
iter = 0;
improvement = Inf;

while iter < maxIter && improvement > epsilon
	oldScore = objectiveFunction(B,x,y,k);
	
	% compute the gradient of f(y)
	delta_y = zeros(n,1);
	for j=1:n
		subvec = exp(-k * (x * y(j) - 1/2));
		vec = (1 + subvec).^(-1);
		
		v1 = 2 * ( B(:,j) - vec ) .* (vec.^2);
		v2 = (-k) * subvec .* x;

		delta_y(j) = dot(v1,v2);
%		for i=1:m
%			a1 = exp(-k * (x(i)*y(j) - 1/2));
%			a2 = (1 + a1)^(-1);
%
%			delta_y(j) = delta_y(j) - 2*(B(i,j) - a2) * a2^2 * a1 * k * x(i);
%		end
	end

	% update the candidate of x
	step = determineStepSizeY(y,delta_y,s,B,x,k);
	y_cand = y - step*delta_y;
	y_cand = max(y_cand, MinValue); % make sure y does not become negative

	newScore = objectiveFunction(B,x,y_cand,k);

	improvement = oldScore - newScore;
	iter = iter + 1;

	if improvement > epsilon
		y = y_cand;
	end
end

yNew = y;

end

