function [ b, C, permutationBtoCrows, permutationCtoBrows, permutationBtoCcols, permutationCtoBcols] = IsNested(B)
%ISNESTED Checks if the given matrix B is nested.
%	returns the following things:
%		b is true if and only if C is nested
%		C is the directly nested version of B
%		permutationBtoCrows - contains the permutation that one needs to apply to the rows of B to obtain C
%		permutationCtoBrows - contains the permutation that one needs to apply to the rows of C to obtain B
%		permutationBtoCcols - contains the permutation that one needs to apply to the cols of B to obtain C
%		permutationCtoBcols - contains the permutation that one needs to apply to the cols of C to obtain B
%
%   This code is free to use for scientific purposes provided that the
%   original paper is cited.
%
%   Copyright Pauli Miettinen & Stefan Neumann, 2016

% Pre-init
    b = false;
    C = B;
    permutationBtoCrows = 1:size(B, 1);
    permutationCtoBrows = 1:size(B, 1);
    permutationBtoCcols = 1:size(B, 2);
    permutationCtoBcols = 1:size(B, 2);
    
[b, C, permutationBtoCrows, permutationCtoBrows] = IsNestedRow(B);
if ~b
	return
end

[b, C, permutationBtoCcols, permutationCtoBcols] = IsNestedRow( transpose(C) );
if ~b
	return
end

C = transpose(C);

end

function [b, C, permutationBtoC, permutationCtoB] = IsNestedRow(B)
m = size(B,1);
n = size(B,2);

C = B(1,:);
permutationBtoC = ones(1,1);

b = true;

% add row i of B to matrix C
for i=2:m
	inserted = false;

	% find out where to insert into C (if possible)
	% at each stage we compare B(i,:) with C(j,:)
	for j=1:(i-1)
		diff = B(i,:) - C(j,:);

		% we check if row B(i,:) is a superset of row C(j,:)
		% if this is the case we insert above C(j,:)
		if (max(diff) <= 1 && min(diff) >= 0)
			% if we have a superset insert
			C = [C(1:(j-1), :); B(i,:); C(j:(i-1), :)];
			permutationBtoC = [permutationBtoC(1:(j-1)), i, permutationBtoC(j:(i-1))];
			inserted = true;
			break;
		elseif ~(max(diff) <= 0 && min(diff) >= -1)
			% if it is also not a subset, then the matrix is not nested
			b = false;
			C = 0;
			permutationBtoC = 0;
			permutationCtoB = 0;
			return
		end
	end

	% it might have been a subset of all rows,
	% in this case we insert at the end
	if ~inserted
		C(i,:) = B(i,:);
		permutationBtoC = [permutationBtoC, i];
	end
end

permutationCtoB(permutationBtoC) = 1:m;
end

