/*
 * Nestedness.c
 * 
 * Implements the Bi-Greedy algorithm for the BMNA problem
 * of Mannila and Terzi KDD'2007.
 *
 * Compile in Matlab with:
 * mex CFLAGS="\$CFLAGS -std=c99 -Wall -O3 -march=native" Nestedness.c BinaryHeap.c
 *
 * This code is free to use for scientific purposes provided that the
 * original paper is cited.
 *
 * Copyright Pauli Miettinen & Stefan Neumann, 2016
 */
#include <stdio.h>
#include <limits.h>
#include <stdbool.h>

#include "matrix.h" // from Matlab
#include "mex.h"

#include "BinaryHeap.h"

mwSize mRows, nCols;
short* B;
BinaryHeap* heapster;

unsigned int
matrixEntry(unsigned int row,
			unsigned int col)
{
	return (col * mRows + row);
}

void
matrixEntryBackward(unsigned int entry,
					unsigned int *row,
					unsigned int *col)
{
	*col = entry / mRows;
	*row = entry % mRows;
}

void
initializeCoverages()
{
	// we will walk through the matrix and check inside the inner most loop
	// whether we have the following flip-box:
	//   B(i,j) == 0 	...   B(i,j+k) == 1
	//   ...					...
	// B(i+l,j) == 1	... B(i+l,j+k) == 0
	for (unsigned int i=0; i<mRows; i++) {
		for (unsigned int j=0; j<nCols; j++) {
			for (unsigned int l=1; i+l < mRows; l++) {
				for (unsigned int k=1; j+k < nCols; k++) {
					// check if we have a submatrix of the form
					// 1 . 0
					// . . .
					// 0 . 1
					if ( B[matrixEntry(i,j)] == 1 && B[matrixEntry(i,j+k)] == 0 
						   	&& B[matrixEntry(i+l,j)] == 0 && B[matrixEntry(i+l,j+k)] == 1 ) {
//						printf("found a submatrix from %d,%d to %d,%d\n", (int)i,(int)j,(int)(i+l),(int)(j+k));
						increaseKey(heapster, matrixEntry(i,j));
						increaseKey(heapster, matrixEntry(i+l,j));
						increaseKey(heapster, matrixEntry(i,j+k));
						increaseKey(heapster, matrixEntry(i+l,j+k));
					}

					// check if we have a submatrix of the form
					// 0 . 1
					// . . .
					// 1 . 0
					if ( B[matrixEntry(i,j)] == 0 && B[matrixEntry(i,j+k)] == 1
						   	&& B[matrixEntry(i+l,j)] == 1 && B[matrixEntry(i+l,j+k)] == 0 ) {
//						printf("found a submatrix from %d,%d to %d,%d\n", (int)i,(int)j,(int)(i+l),(int)(j+k));
						increaseKey(heapster, matrixEntry(i,j));
						increaseKey(heapster, matrixEntry(i+l,j));
						increaseKey(heapster, matrixEntry(i,j+k));
						increaseKey(heapster, matrixEntry(i+l,j+k));
					}
				}
			}
		}
	}
}

void
flipEntryAndUpdateCoverages(uint64_t entry)
{
	bool convertedFromZeroToOne = (B[entry] == 0);
//	if (convertedFromZeroToOne) {
//		printf("converting from 0 to 1\n");
//	} else {
//		printf("converting from 1 to 0\n");
//	}

	// flip entry
	B[entry] = 1 - B[entry]; // flips between 1 and 0

	unsigned int i, j;
	matrixEntryBackward(entry, &i, &j);

//	printf("highest coverage at index %d,%d\n\n", (int)i, (int)j);

	// update coverages
	for (uint64_t v=0; v<mRows; v++) {
		for (uint64_t u=0; u<nCols; u++) {
			if (convertedFromZeroToOne) { // we converted 0 to 1
				// we make the case distinction from the paper
				if ( B[matrixEntry(v,u)] == 0 && B[matrixEntry(i,u)] == 1 && B[matrixEntry(v,j)] == 1) {
//					printf("decreasing key %d,%d\n", (int)v,(int)u);
//					printf("decreasing key %d,%d\n", (int)i,(int)u);
//					printf("decreasing key %d,%d\n\n", (int)v,(int)j);
					decreaseKey(heapster, matrixEntry(v,u));
					decreaseKey(heapster, matrixEntry(i,u));
					decreaseKey(heapster, matrixEntry(v,j));
				} else if (B[matrixEntry(v,u)] == 1 && B[matrixEntry(i,u)] == 0 && B[matrixEntry(v,j)] == 0) {
//					printf("increasing key %d,%d\n", (int)v,(int)u);
//					printf("increasing key %d,%d\n", (int)i,(int)u);
//					printf("increasing key %d,%d\n\n", (int)v,(int)j);
					increaseKey(heapster, matrixEntry(v,u));
					increaseKey(heapster, matrixEntry(i,u));
					increaseKey(heapster, matrixEntry(v,j));
				}
			} else { // we converted 1 to 0
				// we make the case distinction from the paper
				if ( B[matrixEntry(v,u)] == 0 && B[matrixEntry(i,u)] == 1 && B[matrixEntry(v,j)] == 1) {
//					printf("increasing key %d,%d\n", (int)v,(int)u);
//					printf("increasing key %d,%d\n", (int)i,(int)u);
//					printf("increasing key %d,%d\n\n", (int)v,(int)j);
					increaseKey(heapster, matrixEntry(v,u));
					increaseKey(heapster, matrixEntry(i,u));
					increaseKey(heapster, matrixEntry(v,j));
				} else if (B[matrixEntry(v,u)] == 1 && B[matrixEntry(i,u)] == 0 && B[matrixEntry(v,j)] == 0) {
//					printf("decreasing key %d,%d\n", (int)v,(int)u);
//					printf("decreasing key %d,%d\n", (int)i,(int)u);
//					printf("decreasing key %d,%d\n\n", (int)v,(int)j);
					decreaseKey(heapster, matrixEntry(v,u));
					decreaseKey(heapster, matrixEntry(i,u));
					decreaseKey(heapster, matrixEntry(v,j));
				}
			}
		}
	}
}

/**
 * The input matrix B must be provided in prhs, the output will be stored in plhs.
 *
 * This function is inteded to be called by Matlab.
 */
void
mexFunction(const int nlhs,
			mxArray *plhs[],
			const int nrhs,
			const mxArray *prhs[])
{
	const mwSize *dim = mxGetDimensions(prhs[0]);
	mRows = dim[0];
	nCols = dim[1];
	double *C = mxGetPr(prhs[0]);

	// copy the matrix from matlab to our own memory
	// so we won't perform changes that go to matlab
	B = malloc(sizeof(short) * mRows * nCols);
	for (uint64_t k=0; k<mRows*nCols; k++) {
		B[k] = C[k];
	}

	heapster = createBinaryHeap(mRows * nCols);
	for (uint64_t k=0; k < mRows*nCols; k++) {
		insert(heapster, 0, k);
	}

	initializeCoverages();

	unsigned int flippedEntries = 0;

	// while there is some element with positive coversage
	while ( peakMax(heapster) != 0 ) {
		flippedEntries++;
		HeapElement max = deleteMax(heapster);
//		printf("score was: %d\n", (int)max.score);
		
		flipEntryAndUpdateCoverages( max.column );
	}

	freeBinaryHeap(heapster);

	// generate output
	plhs[0] = mxCreateDoubleScalar( flippedEntries );
	

	// generate output of B
	mwSize outDims[2];
	outDims[0] = mRows;
	outDims[1] = nCols;
	mwSize numberOfOutFields = mRows*nCols;

	plhs[1] = mxCreateLogicalArray(2, outDims);
	mxLogical *pn = mxGetLogicals(plhs[1]);
	for (mwSize index=0; index<numberOfOutFields; index++) {
		pn[index] = B[index];
	}
	
	free(B);
}

