function [ r, X, iters ] = RoundingRankLPCA( B, varargin )
%RoundingRankLPCA Computes the rounding rank using LPCA
%   r = RoundingRankLPCA(B) computes the rounding rank of B.
%   [r, X] = RoundingRankLPCA(B) returns also the witness matrix X.
%   RoundingRankLPCA(..., 'param1', val1, ...) passes extra parameters to
%   LPCA_FIT.
%
%   RoundingRankLPCA assumes the rounding threshold is 0.5.
%
%   See also: LPCA_FIT, RoundingRankProj, RoundingRankNuclear, Round
%
%   This code is free to use for scientific purposes provided that the
%   original paper is cited.
%
%   Copyright Pauli Miettinen & Stefan Neumann, 2016

%% Check B is valid
p = inputParser;
addRequired(p, 'B', @(x) validateattributes(x, {'numeric'}, {'2d', 'binary'}, mfilename, 'B', 1));

parse(p, B);

r = Inf;
X = [];

%% Compute
[n, m] = size(B);
N = min(n, m);

for i=1:N
    [U, V, iters] = lpca_fit(B, i, varargin{:});
    A = Round(U*V, 0.5);
    if norm(B - A, 'fro') == 0
        r = i;
        X = U*V;
        break;
    end
end

if r == Inf
    % Let's do a trivial decomposition
    r = N;
    X = B;
end


end

