function [ r, X ] = RoundingRankNuclearAlt( B, varargin )
%RoundingRankNuclear Computes the rounding rank using nuclear norm
%   r = RoundingRankNuclear(B) returns the rounding rank of B (an upper
%       bound thereof)
%   [r, X] = RoundingRankNuclear(B) returns also the rank-r matrix X such
%       that Round(X, 0) = B
%   r = RoundingRankNuclear(B, 'threshold', tau) uses threshold "tau" as the 
%       rounding threshold, i.e. Round(X, tau) = B (default = 0.5)
%   r = RoundingRankNuclear(B, 'solver', s) uses solver "s" in CVX
%   r = RoundingRankNuclear(B, 'verbose', v) sets the verbosity: 0 or false
%       means no output, 1 or true means output (default = true)
%
%   See also: Round, RoundingRankProj, RoundingRankSVD
%
%   This code is free to use for scientific purposes provided that the
%   original paper is cited.
%
%   Copyright Pauli Miettinen & Stefan Neumann, 2016

%% Check input
p = inputParser;
defaultThreshold = 0.5;
defaultSolver = 'sdpt3';
defaultVerbose = true;

addRequired(p, 'B', ...
    @(x) validateattributes(x, {'numeric'}, {'nonempty','2d','binary'}));
addParameter(p, 'threshold', defaultThreshold, @isscalar);
addParameter(p, 'solver', defaultSolver, @ischar);
addParameter(p, 'verbose', defaultVerbose, @isscalar);

parse(p, B, varargin{:});

tau = p.Results.threshold;
solver = p.Results.solver;
verbose = logical(p.Results.verbose);


[N, M] = size(B);

%% CVX
largeNum = 10^6;
epsilon = N*M*eps;

UB = largeNum*ones(size(B));
UB(B == 0) = tau - epsilon;

LB = -largeNum*ones(size(B));
LB(B == 1) = tau + epsilon;

if verbose
    cvx_begin
else
    cvx_begin quiet
end
    cvx_solver(solver);
    variable X(N, M);
    minimize norm_nuc(X);
    subject to
      X <= UB;
      X >= LB;
    %X(B == 0) <= tau;
    %X(B == 1) >= tau;
cvx_end

%% Results
if exist('X', 'var') ~= 1 || isempty(X),
    % CVX didn't found any solution
    if verbose
        fprintf('CVX didn''t found any solution\n');
    end
    X = B;
end
    
r = rank(X);
if verbose
    fprintf('Rank after CVX is %i\n', r);
end

if r < 2
    % Don't run the SVD clearance for rank-1 (or rank-0) matrices
    return
end

if verbose
    fprintf('Computing the SVD...\n');
end


% Cut off the useless part...
if r < min(N, M)/2
    [U, S, V] = svds(X, r);
    % svds is actually slow if r is close full rank
else
    [U, S, V] = svd(X);
    U = U(:,1:r);
    S = S(1:r, 1:r);
    V = V(:, 1:r);
end

if verbose
    fprintf('SVD Done!\n');
end
for k=(r-1):-1:1
    Uk = U(:,1:k);
    Sk = S(1:k, 1:k);
    Vk = V(:, 1:k);
    if norm(B - Round(Uk*Sk*Vk', tau)) > 0
        break;
    end
end

r = k+1;
if verbose
    fprintf('Final rounding rank is %i\n', r);
end
U = U(:,1:r);
S = S(1:r, 1:r);
V = V(:, 1:r);
X = U*S*V';


end

