function [ r, L, R, err ] = RoundingRankProj( B, varargin )
%ROUNDINGRANKPROJ Returns a rounding rank factorisation of B.
%   r = RoundingRankProj(B) returns the rounding rank of B (an upper
%       bound thereof)
%   [r, X] = RoundingRankProj(B) returns also the rank-r matrix X such
%       that Round(X, 0.5) = B
%   [r, L, R] = RoundingRankProj(B) returns the factor matrices L and R
%       such that Round(L*R, 0.5) = B
%   [..., err] = RoundingRankProj(...) returns also the error (0 unless the
%       'minerr' parameter is given)
%   r = RoundingRankProj(B, 'PARAM1', val1, 'PARAM2', val2, ...) specifies
%       optional parameter name/value pairs. Parameters are:
%   'threshold' - The threshold used in the rounding (default = 0.5).
%   'startdim' - The smallest dimensionality to check (default = 1)
%   'minerr' - Find minimum-error fixed rounding rank decomposition. The
%       rank of the decomposition is the value for the parameter. Giving
%       value 0 will disable the feature (default).
%   'replicates' - The number of times the random projection is done
%       (default = 1).
%   'projection' - The method used for computing the random projection.
%       Choises are:
%           'sparse'    - The default. Uses a sparse {-1, 0, 1} matrix.
%           'normal'    - Samples the matrix from the multivariate normal
%                         distribution.
%           'binary'    - Samples a {-1, 1} matrix.
%           'uniform'   - Samples u.a.r. from the unit circle.
%   'solver' - The solver used to solve the linear program. Choises are:
%           'linprog'   - The default. Uses Optimization Toolbox.
%           'gurobi'    - Uses the Gurobi solver directly.
%           'cvx_sedumi'- Uses the SeDuMi solver of the CVX package.
%           'cvx_sdpt3' - Uses the SDPT3 solver of the CVX package.
%           'cvx_gurobi'- Uses the Gurobi solver via the CVX package.
%           'cvx_mosek' - Uses the MOSEK solver via the CVXC package.
%   'verbose' - Set to true makes the algorithm produce intermediate output
%       (default = true). 
%
%   See also: Round, RoundingRankNuclear, RoundingRankSVD, linprog
%
%   This code is free to use for scientific purposes provided that the
%   original paper is cited.
%
%   Copyright Pauli Miettinen & Stefan Neumann, 2016


%% Check input
p = inputParser;

% Defaults
defaultThreshold = 0.5;
defaultReplicates = 1;
defaultDim = 1;
defaultProjection = 'sparse';
defaultSolver = 'linprog';
defaultVerbose = true;
defaultMinErr = 0;

% Valid options
validProjections = {'sparse', 'normal', 'binary', 'uniform'};
validSolvers = {'linprog', 'gurobi', 'cvx_sedumi', 'cvx_sdpt3', 'cvx_gurobi', 'cvx_mosek'};

% Build parser
addRequired(p, 'B', ...
    @(x) validateattributes(x, {'numeric'}, {'nonempty','2d','binary'}, mfilename, 'B', 1));
addParameter(p, 'threshold', defaultThreshold, @isscalar);
addParameter(p, 'replicates', defaultReplicates, ...
    @(x) validateattributes(x, {'numeric'}, {'scalar', 'nonnegative', 'integer'}, mfilename, 'replicates'));
addParameter(p, 'startdim', defaultDim, ...
    @(x) validateattributes(x, {'numeric'}, {'scalar', 'nonnegative', 'integer'}, mfilename, 'startdim'));
addParameter(p, 'minerr', defaultMinErr, ...
    @(x) validateattributes(x, {'numeric'}, {'scalar', 'nonnegative', 'integer'}, mfilename, 'minerr'));
addParameter(p, 'solver', defaultSolver, @ischar);
addParameter(p, 'projection', defaultProjection, @ischar);
addParameter(p, 'verbose', defaultVerbose, @isscalar);

% Parse
parse(p, B, varargin{:});

threshold = p.Results.threshold;
replicates = p.Results.replicates;
startdim = p.Results.startdim;
solver = validatestring(p.Results.solver, validSolvers, mfilename, 'solver');
projection = validatestring(p.Results.projection, validProjections, mfilename, 'projection');
verbose = logical(p.Results.verbose);

searchMinErr = false;
minErrRank = 0;
bestErr = Inf;
bestL = [];
bestR = [];
if p.Results.minerr > 0
    searchMinErr = true;
    minErrRank = p.Results.minerr;
end
    

m = size(B,1);
n = size(B,2);

epsilon = 0.000001;

d = startdim;

if searchMinErr
    d = minErrRank;
end

while d < min(m, n)
    found_solution = false;
    for run = 1:replicates
        verbprint(verbose, 'Dimension %3i, replicate %2i (of %i)\n', d, run, replicates);
        % get m points in R^d in general position
        % store these points in the rows of matrix V
        % (the points are in general position with
        %  probability 1)
        
        verbprint(verbose, '\tDoing projection\n');
        V = getpm(n, d, projection);
        L = B * V;
        
        % the matrix R with the hyperplanes
        R = zeros(d,n);
        failed = false;
        
        verbprint(verbose, '\tFinding the RHS matrix\n');
        for j = 1:n
            b = full(B(:,j));
            b = b * (- threshold - epsilon);
            b( B(:,j) == 0 ) = threshold - epsilon;
            
            Lprime = L;
            Lprime( B(:,j) == 1 , : ) = -1 * Lprime( B(:,j) == 1 , : );
            
            if ~searchMinErr
                [x, exitflag] = solvelp(Lprime, b, solver);
                % check the correctness of the solution
                if (exitflag == 1) % | (exitflag == 0) % success
                    % set j'th row of R
                    R(:,j) = x;
                else
                    failed = true;
                    break;
                end
            else
                [x, e, exitflag] = solveminerr(Lprime, b, solver);
                R(:,j) = x;
                if exitflag ~= 1
                    warning('Column %i optimization raised exitflag %i', j, exitflag);
                end
            end
            
            %[x,~,exitflag] = linprog_gurobi(zeros(d,1), Lprime, b);
            

        end
        if failed
            continue;
        end
        if ~searchMinErr
            if isequal(B, Round(L*R, threshold))
                found_solution = true;
                break;
            else
                warning('Found a feasible solution that does not admit the exact solution, but causes error %f', norm(B - Round(L*R, threshold)));
            end
        else
            err = sum(sum(abs(B - Round(L*R, threshold))));
            found_solution = true;
            if err < bestErr
                bestErr = err;
                bestL = L;
                bestR = R;
            end
        end
    end
    if found_solution
        break;
    end
    d = d+1;
end

verbprint(verbose, 'Finished looking for solution\n');

if ~found_solution 
    verbprint(verbose, 'No solution found, returning the trivial one\n');
    % No solution was found, return the trivial one
    r = min(m, n);
    if r == m
        L = eye(m);
        R = B;
    else
        L = B;
        R = eye(n);
    end
else
    if ~searchMinErr
        r = d;
        verbprint(verbose, 'Found solution with rank = %i\n', r);
        % L and R are already correct
    else
        r = d;
        err = bestErr;
        L = bestL;
        R = bestR;
        verbprint(verbose, 'Found answer with error = %d\n', err);
    end
end

if nargout == 2
    L = L*R;
end

end

%% verbprint - print based on verbosity
function nbytes = verbprint(v, varargin)
    nbytes = 0;
    if v
        nbytes=fprintf(1, varargin{:});
    end
end

%% getpm - get the projection matrix
function V = getpm(n, d, projection)
switch projection
    case 'normal'
        V = randn(n, d);
    case 'binary'
        V = 2*double(rand(n, d) < 0.5);
        V = V - 1;
    case 'sparse'
        nonz = find(rand(n*d, 1) < 1/3);
        nonz = nonz(randperm(length(nonz)));
        cut = round(length(nonz)/2);
        [plusI, plusJ] = ind2sub([n, d], nonz(1:cut));
        [minusI, minusJ] = ind2sub([n, d], nonz((cut+1):end));
        I = [plusI; minusI];
        J = [plusJ; minusJ];
        V = [sqrt(3)*ones(length(plusI),1); -sqrt(3)*ones(length(minusI), 1)];
        V = sparse(I, J, V, n, d);
    case 'uniform'
        V = rand(n,d) - 1/2;
        V = normr(V);
    otherwise
        error('The projection method %s is not known', projection);
end
        
end % function getpm

%% solvelp - solves the linear program
function [x, exitflag] = solvelp(A, b, solver)
d = size(A, 2);
switch solver
    case 'linprog'
        options = optimoptions('linprog', 'Display', 'off');
        [x, ~, exitflag] = linprog(zeros(d,1), A, b, [], [], [], [], [], options);
    case 'gurobi'
        [x,~,exitflag] = linprog_gurobi(zeros(d,1), A, b);
    case {'cvx_sedumi', 'cvx_sdpt3', 'cvx_gurobi', 'cvx_mosek'}
        alg = solver(5:end); % The name of the solver
        cvx_begin quiet
          cvx_solver(alg);
          variable x(d);
          subject to 
            A*x <= b;
        cvx_end
        if cvx_optval == 0
            exitflag = 1; % constraints are feasible
        else
            exitflag = 2; 
            x = zeros(d, 1);
        end
    otherwise
        error('Unknown solver %s', solver);
end
end

%% solveminerr - solvest the minimum-margin LP
function [x, val, exitflag] = solveminerr(A, b, solver)
[m, d] = size(A);
A = [A -1*eye(m)]; % append -1 multipliers for the margin
switch solver
    case 'linprog'
        options = optimoptions('linprog', 'Display', 'off');
        f = [zeros(d, 1); ones(m, 1)];
        lb = [-inf(d, 1); zeros(m, 1)]; % Bound c(i) >= 0
        [x, val, exitflag] = linprog(f, A, b, [], [], lb, [], [], options);
        x = x(1:d);
    otherwise
        error('Cannot use solver %s to solve the minimum-margin problem', solver);
end
end
