function [r,X] = RoundingRankSVD(B, varargin)
%RoundingRankSVD Outputs a rounding rank decomposition using svd.
%   r = RoundingRankSVD(B) returns the rounding rank of r (an upper bound
%   thereof)
%   [r, X] = RoundingRankSVD(B) returns also the rank-r matrix X such
%       that Round(X, 0) = B
%   r = RoundingRankSVD(B, 'threshold', tau) uses threshold "tau" as the 
%       rounding threshold, i.e. Round(X, tau) = B (default = 0.5)
%   r = RoundingRankSVD(B, 'verbose', v) sets the verbosity: 0 or false
%       means no output, 1 or true means output (default = true)
%
%   See also: Round, RoundingRank, RoundingRankNuclear
%
%   This code is free to use for scientific purposes provided that the
%   original paper is cited.
%
%   Copyright Pauli Miettinen & Stefan Neumann, 2016

%% Check input
p = inputParser;
defaultThreshold = 0.5;
defaultVerbose = true;

addRequired(p, 'B', ...
    @(x) validateattributes(x, {'numeric'}, {'nonempty','2d','binary'}));
addParameter(p, 'threshold', defaultThreshold, @isscalar);
addParameter(p, 'verbose', defaultVerbose, @isscalar);

parse(p, B, varargin{:});

tau = p.Results.threshold;
verbose = logical(p.Results.verbose);

m = size(B,1);
n = size(B,2);

%% Do the math
if verbose
    fprintf('Computing the SVD...\n');
end
[U,S,V] = svd(full(B));

if verbose
    fprintf('Done!\nTrying different ranks...\n');
end
for i = 1:min(m,n)
    if verbose
        fprintf('\tr = %i\n', i);
    end
	L = U(:, 1:i) * S(1:i,1:i);
	R = V(:,1:i)';

	if isequal(Round(L*R, tau), B)
		break;
	end
end
r = i;
if verbose
    fprintf('Done!\nRounding rank is %i\n', r);
end
X = U(:, 1:r)*S(1:r, 1:r)*V(:, 1:r)';
end

