function [ r ] = Shays( B )
%SHAYS Runs Shay's algorithm for estimating the rounding rank
%
%   SHAYS(B) returns an upper bound of the rounding rank of B.
%
%   SHAYS implements the 'Permutation' algorithm from the main paper. It is
%   originally described in N. Alon, S. Moran, and A. Yehudayoff. Sign rank
%   versus VC dimension. In COLT, pages 47-80, 2016.
%
%   This code is free to use for scientific purposes provided that the
%   original paper is cited.
%
%   Copyright Pauli Miettinen & Stefan Neumann, 2016

m = size(B,1);
n = size(B,2);

if m ~= n
	error('Only works for quadratic matrices.');
end

% Welzl's algorithm
p_current = 1/n * ones(1,n);
p_new = ones(1,n);

F = zeros(n,n);

% find rows and columns of the one-entries in B
[U,V] = find(B);

for i = 1:(n-1)
	% find edge e to pick
	uBest = n+1;
	vBest = n+1;
	weightBest = Inf;

	for u = 1:n
		for v = (u+1):n
			% e is the edge from u to v

			% check if e is already in F:
			if F(u,v) == 1
				continue
			end

			% check if e adds a cycle to F:
			%% (we do this by using the following theorem:
			%%	G is acyclic if and only if 0.5*trace(L)=rank(L),
			%%  where G is a graph and L is the Laplacian of the
			%%  adjacency matrix of G.)
			Fcopy = F;
			Fcopy(u,v) = 1;
			Fcopy(v,u) = 1;

			Fcopy = -Fcopy;
			for i = 1:n
				Fcopy(i,i) = -sum(Fcopy(i,:));
			end

			if 0.5 * trace(Fcopy) ~= rank(Fcopy)
				% there must be a cycle
				continue
			end

			% compute the weight of e:
			weight = 0;
			for j = 1:n
				if B(u,j) ~= B(v,j)
					weight = weight + p_current(j);
				end
			end

			if weight < weightBest
				uBest = u;
				vBest = v;
				weightBest = weight;
			end
		end
	end

	% add the best edge to F
%	disp(strcat('adding edge ', num2str(uBest-1), ' -> ', num2str(vBest-1), ' with weight ', num2str(weightBest)));
	F(uBest,vBest) = 1;
	F(vBest,uBest) = 1;

	% update the probability distribution
	for k = 1:n
		if B(uBest,k) ~= B(vBest,k)
			p_new(k) = 2*p_current(k) / (1 + weightBest);
		else
			p_new(k) = p_current(k) / (1 + weightBest);
		end
	end
	p_current = p_new;
end

% use permutation from Welzl's algorithm
[U,V] = find(F);
tour = zeros(1,n);
visited = zeros(1,n);

u = U(1);
v = V(1);

% we start at node u
i = 2;
tour(1) = u;
visited(u) = 1;

% in each step of the loop we will traverse edge u to v
while i <= n
	% go from u to v
	F(u,v) = 0; % delete edge from u to v
	
	if visited(v) == 0
		visited(v) = 1; % mark v as visited
		tour(i) = v; % add v to tour
		i = i+1;
	end

	% go to the next unvisited node (if possible)
	% or back afterwards
	f = find(F(v,:) == 1);
	if (size(f,2) > 0)
		u = v;
		v = f(1);
		F(v,u) = 2; % mark u as predecessor of v
	else
		f = find(F(v,:) == 2);
		u = v;
		v = f(1);
	end
end

%tour-1

% compute permutation C of B
C = zeros(m,n);
for i = 1:m
	C(i,:) = B(tour(i),:);
end

% find maximum number of sign changes in each column of C
maxSignChanges = 0;
for j = 1:n
	signChanges = 0;

	current = C(1,j);
	for i = 2:m
		if C(i,j) ~= current
			current = C(i,j);
			signChanges = signChanges + 1;
		end
	end

	if signChanges > maxSignChanges
		maxSignChanges = signChanges;
	end
end

r = maxSignChanges + 1;

