/*
 * ShaysInC.c
 * 
 * Implements the O(N / log N) approximation algorithm
 * for sign rank.
 *
 * Compile in Matlab with:
 * mex CFLAGS="\$CFLAGS -std=c99 -Wall -O3 -march=native" ShaysInC.c
 *
 *  ShaysInC(B) returns an upper bound of the rounding rank of B.
 *
 *  ShaysInC implements the 'Permutation' algorithm from the main paper. It is
 *  originally described in N. Alon, S. Moran, and A. Yehudayoff. Sign rank
 *  versus VC dimension. In COLT, pages 47-80, 2016.
 *
 *  This code is free to use for scientific purposes provided that the
 *  original paper is cited.
 *
 *  Copyright Pauli Miettinen & Stefan Neumann, 2016
 */
#include <stdio.h>
#include <limits.h>
#include <stdbool.h>

#include "matrix.h" // from Matlab
#include "mex.h"

double* B;

typedef struct {
	unsigned int firstEdge;
	bool wasVisited;
} Vertex;

typedef struct {
	unsigned int nextEdge;

	unsigned int child;
	unsigned int parent;
} Edge;

typedef struct {
	unsigned int numberOfVertices;
	unsigned int numberOfEdges;

	Vertex* vertices;
	Edge* edges;
} Tree;

unsigned int
matrixEntry(unsigned int row,
			unsigned int col,
			unsigned int mRows,
			unsigned int nCols)
{
	return (col * mRows + row);
}

void
setupTree(Tree *t,
		  unsigned int numberOfNodes)
{
	t->numberOfVertices = numberOfNodes;
	t->numberOfEdges = 0;

	for (unsigned int k = 0; k<numberOfNodes; k++) {
		t->vertices[k].firstEdge = UINT_MAX;
		t->vertices[k].wasVisited = false;
	}

	for (unsigned int k = 0; k<numberOfNodes*2; k++) {
		t->edges[k].nextEdge = UINT_MAX;
		t->edges[k].child = UINT_MAX;
		t->edges[k].parent = UINT_MAX;
	}
}

Tree*
initTree(unsigned int numberOfNodes)
{
	Tree* t = mxMalloc(sizeof(Tree));
	t->vertices = mxMalloc( sizeof(Vertex) * numberOfNodes );
	t->edges = mxMalloc( sizeof(Edge) * 2*numberOfNodes );

	setupTree(t, numberOfNodes);

	return t;
}

Tree*
copyTree(Tree *copy,
		 Tree *t)
{
	setupTree(copy, t->numberOfVertices);
	copy->numberOfEdges = t->numberOfEdges;

	for (unsigned int k=0; k < t->numberOfVertices; k++) {
		copy->vertices[k].firstEdge = t->vertices[k].firstEdge;
	}

	for (unsigned int k=0; k < t->numberOfEdges; k++) {
		copy->edges[k].nextEdge = t->edges[k].nextEdge;
		copy->edges[k].child = t->edges[k].child;
		copy->edges[k].parent = t->edges[k].parent;
	}

	return copy;
}

void
freeTree(Tree *t)
{
	mxFree(t->edges);
	mxFree(t->vertices);
	mxFree(t);
}

void
addDirectedEdge(Tree* F,
				unsigned int parent,
				unsigned int child)
{
	F->edges[F->numberOfEdges].parent = parent;
	F->edges[F->numberOfEdges].child = child;

	if (F->vertices[parent].firstEdge == UINT_MAX) {
		F->vertices[parent].firstEdge = F->numberOfEdges;
	} else {
		unsigned int i = F->vertices[parent].firstEdge;
		while (F->edges[i].nextEdge != UINT_MAX) {
			i = F->edges[i].nextEdge;
		}
		F->edges[i].nextEdge = F->numberOfEdges;
	}

	F->numberOfEdges += 1;
}

void
addEdge(Tree* F,
		unsigned int parent,
		unsigned int child)
{
	addDirectedEdge(F,parent,child);
	addDirectedEdge(F,child,parent);
}

bool
containsEdge(Tree* F,
			 unsigned int parent,
			 unsigned int child)
{
	if (F->vertices[parent].firstEdge == UINT_MAX) {
		return false;
	} else {
		unsigned int i = F->vertices[parent].firstEdge;
		do {
			if (F->edges[i].child == child) {
				return true;
			}

			i = F->edges[i].nextEdge;
		} while (i != UINT_MAX);
	}

	return false;
}

bool
containsCycleAtNode(Tree *t,
					unsigned int vertexIndex,
					unsigned int parent)
{
	t->vertices[vertexIndex].wasVisited = true;

	unsigned int currentEdgeIndex = t->vertices[vertexIndex].firstEdge;
	while (currentEdgeIndex != UINT_MAX) {
		unsigned int child = t->edges[currentEdgeIndex].child;

		if (! t->vertices[child].wasVisited) {
			if (containsCycleAtNode(t, child, vertexIndex)) {
				return true;
			}
		} else if (child != parent && parent != UINT_MAX) {
			return true;
		}

		currentEdgeIndex = t->edges[currentEdgeIndex].nextEdge;
	}

	return false;
}

bool
containsCycle(Tree* t)
{
	if (t->numberOfEdges == 0) {
		return false;
	}

	// mark all nodes as not visited
	for (unsigned int i=0; i < t->numberOfVertices; i++) {
		t->vertices[i].wasVisited = false;
	}

	// start cycle detection from each unvisited node
	// (we have to start from each of those as the graph may have
	// 	multiple components)
	for (unsigned int i=0; i < t->numberOfVertices; i++) {
		if (! t->vertices[i].wasVisited) {
			// if we find a cycle at an unvisited node, return true
			if (containsCycleAtNode(t, i, UINT_MAX)) {
				return true;
			}
		}
	}

	return false;
}

void
traverseGraphUtil(Tree* t,
				  unsigned int* tour,
				  unsigned int vertexIndex,
				  unsigned int parent,
				  unsigned int* numberOfVisitedVertices)
{
	// stop once we have seen all vertices
	if (vertexIndex == UINT_MAX) {
		return;
	}

	if (*numberOfVisitedVertices >= t->numberOfVertices) {
		return;
	}

	// mark vertex as visited
	if (! t->vertices[vertexIndex].wasVisited) {
		t->vertices[vertexIndex].wasVisited = true;
		tour[*numberOfVisitedVertices] = vertexIndex;
		*numberOfVisitedVertices += 1;
	}

	unsigned int currentEdgeIndex = t->vertices[vertexIndex].firstEdge;
	while (currentEdgeIndex != UINT_MAX) {
		unsigned int child = t->edges[currentEdgeIndex].child;

		if ( ! t->vertices[child].wasVisited ) {
			traverseGraphUtil(t, tour, child, vertexIndex, numberOfVisitedVertices);
		}

		currentEdgeIndex = t->edges[currentEdgeIndex].nextEdge;
	}
}

void
traverseGraph(Tree* t,
			  unsigned int* tour)
{
	// mark all vertices as not visited
	for (unsigned int i = 0; i < t->numberOfVertices; i++) {
		t->vertices[i].wasVisited = false;
	}

	unsigned int numberOfVisitedVertices = 0;
	traverseGraphUtil(t, tour, 0, UINT_MAX, &numberOfVisitedVertices);
}
			  
/**
 * The input matrix B must be provided in prhs, the output will be stored in plhs.
 *
 * This function is inteded to be called by Matlab.
 */
void
mexFunction(const int nlhs,
			mxArray *plhs[],
			const int nrhs,
			const mxArray *prhs[])
{
	const mwSize *dim = mxGetDimensions(prhs[0]);
	const mwSize mRows = dim[0];
	const mwSize nCols = dim[1];
	B = mxGetPr(prhs[0]);

	double p_current[mRows], p_new[mRows];
	for (unsigned int i = 0; i<mRows; i++) {
		p_current[i] = 1.0 / (double)mRows;
	}

	Tree* F = initTree(mRows);
	Tree* Fcopy = initTree(mRows);

	for (unsigned int i = 0; i < (mRows - 1); i++) {
		unsigned int uBest = UINT_MAX;
		unsigned int vBest = UINT_MAX;
		double weightBest = DBL_MAX;

		for (unsigned int u=0; u<mRows; u++) {
			for (unsigned int v=u+1; v<mRows; v++) {
				// check if F already contains (u,v)
				if ( containsEdge(F, u, v) ) {
					continue;
				}

				// check if F has a cycle with (u,v)
				copyTree(Fcopy,F);
				addEdge(Fcopy,u,v);
				if (containsCycle(Fcopy)) {
					continue;
				}

				// compute the weight of e
				double weight = 0;
				for (unsigned int j = 0; j<mRows; j++) {
					double m1 = B[matrixEntry(u,j,mRows,nCols)];
					double m2 = B[matrixEntry(v,j,mRows,nCols)];

					if (m1 != m2) {
						weight += p_current[j];
					}
				}

				if (weight < weightBest) {
					uBest = u;
					vBest = v;
					weightBest = weight;
				}
			}
		}

		addEdge(F,uBest,vBest);

		// update the probability distributions
		for (unsigned int k=0; k<mRows; k++) {
			unsigned int m1 = B[matrixEntry(uBest,k,mRows,nCols)];
			unsigned int m2 = B[matrixEntry(vBest,k,mRows,nCols)];

			if (m1 != m2) {
				p_new[k] = 2.0*p_current[k] / (1.0 + weightBest);
			} else {
				p_new[k] = p_current[k] / (1.0 + weightBest);
			}
		}

		for (unsigned int k=0; k<mRows; k++) {
			p_current[k] = p_new[k];
		}
	}

	// traverse F in order to get a tour
	unsigned int tour[mRows];
	traverseGraph(F, tour);

	freeTree(Fcopy);
	freeTree(F);

	// use the tour in order to get a permutation of B
	bool C[mRows][nCols];
	for (unsigned int i = 0; i<mRows; i++) {
		for (unsigned int j = 0; j<nCols; j++) {
			C[i][j] = B[matrixEntry(tour[i],j,mRows,nCols)];
		}
	}

	// find maximum number of sign changes in each column of C
	unsigned int maxSignChanges = 0;
	for (unsigned int j = 0; j<nCols; j++) {
		unsigned int signChanges = 0;
		bool currentSign = C[0][j];

		for (unsigned int i = 1; i<mRows; i++) {
			if ( currentSign != C[i][j] ) {
				signChanges += 1;
				currentSign = C[i][j];
			}
		}

		if (signChanges > maxSignChanges) {
			maxSignChanges = signChanges;
		}
	}

	// generate output
	plhs[0] = mxCreateDoubleScalar( maxSignChanges + 1 );
}

