function [ A, L, R, iter ] = StochasticGradientDescent(B,r,varargin)%,k,s,maxIter,epsilon,x_initial,y_initial)
%StochasticGradientDescent Heuristically computes the closest matrix of B of rounding rank r.
% as for the optional parameters one can pass the following things (in this order):
%	r	-	the size of the factorizations
%	k	-	the steepness of the logistic function
%	s	-	the minimal stepsize for the gradient descent
%	maxIter	-	maximum number of iterations
%	epsilon	-	threshold after which we stop improving our solution
%	L_initial	-	initial solution for gradient descent for L
%	R_initial	-	initial solution for gradient descent for R
%
%   This code is free to use for scientific purposes provided that the
%   original paper is cited.
%
%   Copyright Pauli Miettinen & Stefan Neumann, 2016


m = size(B,1);
n = size(B,2);

if nargin < 3
	k = 2; % steepness of the logistic function
else
	k = varargin{1};
end
if nargin < 4
	s = max(m,n)^(-1/2); % stepsize for the gradient descent
else
	s = varargin{2};
end
if nargin < 5
	maxIter = m*n;
else
	maxIter = varargin{3};
end
if nargin < 6
	epsilon = 1; % we will stop once the error decreases by less than epsilon
else
	epsilon = varargin{4};
end
if nargin < 7
	L_initial = rand(m,r) + sqrt(0.5 / r) - 0.5;
else
	L_initial = varargin{5};
end
if nargin < 8
	R_initial = rand(r,n) + sqrt(0.5 / r) - 0.5;
else
	R_initial = varargin{6};
end

iter = 0;
improvement = Inf;

L = zeros(m,r);
R = zeros(r,n);
L_cand = L_initial;
R_cand = R_initial;

oldScore = objectiveFunction(B,L,R,k);
initialScore = oldScore;

entries = randperm(m*n) - 1;

%for entry = 1:(m*n)
while iter < maxIter
    %if mod(iter,100) == 0
    %		iter
    %	end

	% matrix entry we are optimizing
	mismatchedEntries = find(B - Round(L*R));
	entry = mismatchedEntries( randi(length(mismatchedEntries)) );
	row = floor((entry-1) / n) + 1;
	if row == 0
		row = 1;
	end
	col = mod(entry,n) + 1;
%	row = floor(entries(entry) / n) + 1;
%	col = mod(entries(entry),n) + 1;

	L = L_cand;
	R = R_cand;

	oldScore = objectiveFunction(B,L,R,k);

%	disp('improving L');
	L_cand = improveL(B, r, L, R, 1, epsilon, k, s, m, n, row);
%	disp('improving R');
	R_cand = improveR(B, r, L, R, 1, epsilon, k, s, m, n, col);

	newScore = objectiveFunction(B,L_cand,R_cand,k);

	improvement = oldScore - newScore;
	iter = iter + 1;
end

initialScore;
oldScore = objectiveFunction(B,L,R,k);

A = L * R;
end

function [ f ] = objectiveFunction(B, L, R, k)
f = norm( B - (1 + exp( -k * (L*R - 1/2) )).^(-1), 'fro' )^2;
end

function [ LNew, scoreNew ] = improveL(B, r, L, R, maxIter, epsilon, k, s, m, n, row)
iter = 0;
improvement = Inf;

while iter < maxIter && improvement > epsilon
	oldScore = objectiveFunction(B,L,R,k);
	
	% compute the gradient of f(x)
	delta_L = zeros(1,r);
	for j=1:n
		a1 = exp(-k * (dot(L(row,:),R(:,j)) - 1/2) );
		a2 = 1 / (1+a1);

		delta_L = delta_L - 2*(B(row,j) - a2) * a2^2 * a1 * k * R(:,j)';
	end

	% update the candidate of x
	step = determineStepSizeL(L, delta_L, s, B, R, k, row);
	L_cand = L;
	L_cand(row,:) = L_cand(row,:) - step*delta_L;

	newScore = objectiveFunction(B,L_cand,R,k);

	improvement = oldScore - newScore;
	iter = iter + 1;

	if improvement > epsilon
		L = L_cand;
	end
end

LNew = L;
end

function [ step ] = determineStepSizeL(L, delta_L, s, B, R, k, row)
step = s;
curScore = objectiveFunction(B,L,R,k);
newScore = 0;

L_cand = L;
while curScore > newScore
	curScore = newScore;

	step = 2*step;
	L_cand(row,:) = L(row,:) - step*delta_L;

	newScore = objectiveFunction(B,L_cand,R,k);
end

step = step/2;
end

function [ RNew, newScore ] = improveR(B, r, L, R, maxIter, epsilon, k, s, m, n, col)
iter = 0;
improvement = Inf;

R_cand = R;
while iter < maxIter && improvement > epsilon
	oldScore = objectiveFunction(B,L,R,k);
	
	delta_R = zeros(r,1);
	for i=1:m
		a1 = exp(-k * (dot(L(i,:),R(:,col)') - 1/2));
		a2 = 1 / (1+a1);

		delta_R = delta_R - 2 * (B(i,col) - a2) * a2^2 * a1 * k * L(i,:)';
	end

	% update the candidate of R
	step = determineStepSizeR(R,delta_R,s,B,L,k,col);
	R_cand(:,col) = R(:,col) - step*delta_R;

	newScore = objectiveFunction(B,L,R_cand,k);

	improvement = oldScore - newScore;
	iter = iter + 1;

	if improvement > epsilon
		R = R_cand;
	end
end

RNew = R;
end

function [ step ] = determineStepSizeR(R, delta_R, s, B, L, k, col)
step = s;
curScore = objectiveFunction(B,L,R,k);
newScore = 0;

R_cand = R;
while curScore > newScore
	curScore = newScore;

	step = 2*step;
	R_cand(:,col) = R(:,col) - step*delta_R;

	newScore = objectiveFunction(B,L,R_cand,k);
end

step = step/2;
end

