function [N, Norig] = createNestedMatrix(rows, cols, dens, noise)
%createNestedMatrix creates a nested matrix of given parameters
%  N = createNestedMatrix(rows, cols, dens, noise) returns a rows-by-cols
%  nested matrix with density dens (on expectation) and noise level noise
%  (w.r.t. number of 1s in the original data).
%
%  [N, Norig] = createNestedMatrix(...) will also return the noise-free
%  matrix Norig.
%
%  createNestedMatrix uses the exponential distribution for the
%  nonnegative factor matrices. 

p = sqrt(dens); % The expected value

a = randexp(p, rows, 1);
b = randexp(p, 1, cols);

Norig = Round(a*b, 0.5);

if noise > 0
    N = double(xor(Norig, rand(rows, cols) < noise*nnz(Norig)/(rows*cols)));
else
    N = Norig;
end

end

function x = randexp(mu, varargin)
%RANDEXP samples random numbers from the exponential distribution
%  x = RANDEXP(mu) returns exponentially-distributed scalar x with
%  expected value mu
%  x = RANDEXP(mu, n, m) returns an n-by-m array of random values
    
    defaultN = 1;
    defaultM = 1;
    p = inputParser;
    addRequired(p, 'mu', @(x) validateattributes(x, {'numeric'}, {'real', ...
                        'scalar', 'nonnegative'}, 'randexp', 'mu', 1));
    addOptional(p, 'n', defaultN, ...
                @(x) validateattributes(x, {'numeric'}, {'integer', ...
                        'scalar', 'positive'}, 'randexp', 'n', 2));
    addOptional(p, 'm', defaultM, ...
                @(x) validateattributes(x, {'numeric'}, {'integer', ...
                        'scalar', 'positive'}, 'randexp', 'm', 3));
    
    parse(p, mu, varargin{:});
    n = p.Results.n;
    m = p.Results.m;
    
    u = rand(n, m);
    
    x = -mu*log(1 - u); % The inverse cumulative function of the
                        % exponential distribution
end