function [ f, A, b, lb ] = init_proj_lp( B, L, threshold, epsilon )
%INIT_PROJ_LP Returns the LP for the projection method
%   The function has two modes, decided on the requested output parameters.
%   In the first mode, only two parameters are requested, in which case the
%   function returns the design matrix A and bounds b for the exact
%   decomposition. In the second mode, four parameters are requested, and
%   the function returns also the evaluation function f, and lower bounds
%   lb for the minimum-error fixed rank LP.
%
%   Usage:
%   [A, b] = INIT_PROJ_LP(B, L, threshold, epsilon) returns the design matrix
%   A and bound vector b for the exact decomposition. B is a column of the 
%   original binary matrix, L is the left-hand factor matrix, threshold is the rounding threshold, and 
%   epsilon is the margin.
%
%   [f, A, b, lb] = INIT_PROJ_LP(B, L, threshold, epsilon) returns the
%   objective function f, design matrix A, bound vector b, and lower bounds
%   lb for the least-error fixed rounding rank LP.
%
%   See also:
%   LINPROG
%
%   This code is free to use for scientific purposes provided that the
%   original paper is cited.
%
%   Copyright Pauli Miettinen & Stefan Neumann, 2016

if ~(nargout == 2 || nargout == 4)
    error('The algorithm must output exactly 2 or 4 parameters');
end

if ~iscolumn(B) || ~ismatrix(L) || ~isscalar(threshold) || ~isscalar(epsilon)
    error('The first parameter must be a column vector and the other two a scalar');
end
    

b = B;
b = b * (- threshold - epsilon);
b( B == 0 ) = threshold - epsilon;
A = L;
A( B == 1 , : ) = -1 * A( B == 1 , : );

if nargout == 2
    % Make A and b be the two first output variables
    f = A;
    A = b;
    return;
end

% Add the margin variables
[m, d] = size(A);
A = [A -1*eye(m)];
f = [zeros(d, 1); ones(m, 1)]; % Minimize the L1 of c
lb = [-inf(d, 1); zeros(m, 1)]; % Bound c(i) >= 0


end

