function [bestErr, bestL, bestR, iter] = iterNestedNNRank1(A, tau, maxiter, R, L)
%iterNestedNNRank1 finds a nested matrix that is close to the input
%   [err, L, R, iter] = ITERNESTEDNNRANK1(A, tau, maxiter, iR, iL) finds the
%   nested matrix closest to A using rounding threshold tau, maximum number
%   of iterations maxiter, and initial starting factor matrices iR and iL.
%   It will return the distance in err, the factor matrices in L and R, and
%   the number of iterations needed in iter. 
%
%   [] = ITERNESTEDNNRANK1(A) will use tau=0.5, maxiter=300, and initialize
%   L and R appropriately.
%
%   This code is free to use for scientific purposes provided that the
%   original paper is cited.
%
%   Copyright Pauli Miettinen & Stefan Neumann, 2016

if nargin < 2
    tau = 0.5;
end

if nargin < 3
    maxiter = 300;
end

[n, m] = size(A);

%R = rand(1, m);
if nargin < 4
  s = sum(A);
  [~, o] = sort(s, 'descend');
  I(o) = 1:m; % inverse permutation
  x = (1:m).^(-1);
  R = x(I);
  %R = sum(A)/n;
end

if nargin < 5
  L = nan( n, 1); % pre-allocate
end

bestErr = nnz(A);
bestL = zeros(n, 1);
bestR = zeros(1, m);

for iter=1:maxiter
    %% Update L
    for i = 1:n
        a = A(i, :);
        I = find(a);
        vals = R(I);
        [vals, ord] = sort(vals, 'descend'); % descending order
        if ~isnan(L(i))
	  oldErr = nnz(xor(a, Round(L(i)*R, tau)));
	  bestX = L(i);
        else
          oldErr = Inf;
        end
        zeroErr = nnz(a);
        zeroX = max(eps, tau/max(R) - 10*eps); % Make sure everything
                                               % rounds to 0
	%DEBUG
	%fprintf('L: oldErr = %i, zeroErr = %i\n', oldErr, zeroErr);
	if zeroErr < oldErr
	  oldErr = zeroErr;
          bestX = zeroX;
        end
        for j=1:length(vals)
            x = tau/vals(j) + eps; % make sure x*vals(j) > tau
            newErr = nnz(xor(a, Round(x*R, tau)));
            %DEBUG
	    %fprintf('j = %i, vals(j) = %f, nnz(xR) = %i, err = %i\n', j, vals(j), nnz(Round(x*R, tau)), newErr);
	    %fprintf('L: j = %i, newErr = %i, oldErr = %i\n', j, newErr, oldErr);
            if newErr < oldErr
                oldErr = newErr;
                bestX = x;
            %else
                %break;
            end
        end
        L(i) = bestX;
    end
    %L

    %% Update R
    for i = 1:m
        a = A(:, i);
        I = find(a);
        vals = L(I);
        [vals, ord] = sort(vals, 'descend'); % descending order
        oldErr = nnz(xor(a, Round(L*R(i), tau)));
        bestX = R(i);
        zeroErr = nnz(a);
        zeroX =  max(eps, tau/max(L) - 10*eps);
	%DEBUG
	%fprintf('R: oldErr = %i, zeroErr = %i\n', oldErr, zeroErr);
        if zeroErr < oldErr
          oldErr = zeroErr;
          bestX = zeroX;
        end
        for j=1:length(vals)
            x = tau/vals(j) + eps; % make sure x*vals(j) > tau
            newErr = nnz(xor(a, Round(L*x, tau)));
	    %DEBUG
	    %fprintf('j = %i, vals(j) = %f, nnz(Lx) = %i, err = %i\n', j, vals(j), nnz(Round(L*x, tau)), newErr);
	    %fprintf('R: j = %i, newErr = %i, oldErr = %i\n', j, newErr, oldErr);
            if newErr < oldErr
                oldErr = newErr;
                bestX = x;
            %else
                %break;
            end
        end
        R(i) = bestX;
    end
    %R

    newErr = nnz(xor(A, Round(L*R, tau)));
    if newErr < bestErr
        bestErr = newErr;
        bestL   = L;
        bestR   = R;
    else
        break
    end
end
