function [ L, R, err ] = minErrRRProj( B, k, varargin )
%MINERRRRPROJ Returns a minimum-error rounding rank factorisation of B.
%   [L, R] = minErrRRProj(B, k) returns the rank-k factor matrices s.t. 
%       Round(L,R) is close to B.
%   [L, R, err] = minErrRRProj(B, k) returns also the error between 
%       Round(L,R) and B.
%   minErrRRProj(B, k, 'PARAM1', val1, 'PARAM2', val2, ...) specifies
%       optional parameter name/value pairs. Parameters are:
%   'threshold' - The threshold used in the rounding (default = 0.5).
%   'replicates' - The number of times the random projection is done
%       (default = 1).
%   'projection' - The method used for computing the random projection.
%       Choises are:
%           'sparse'    - The default. Uses a sparse {-1, 0, 1} matrix.
%           'normal'    - Samples the matrix from the multivariate normal
%                         distribution.
%           'binary'    - Samples a {-1, 1} matrix.
%           'uniform'   - Samples u.a.r. from the unit circle.
%   'solver' - The solver used to solve the linear program. Choises are:
%           'linprog'   - The default. Uses Optimization Toolbox.
%           'gurobi'    - Uses the Gurobi solver directly.
%           'cvx_sedumi'- Uses the SeDuMi solver of the CVX package.
%           'cvx_sdpt3' - Uses the SDPT3 solver of the CVX package.
%           'cvx_gurobi'- Uses the Gurobi solver via the CVX package.
%           'cvx_mosek' - Uses the MOSEK solver via the CVXC package.
%   'verbose' - Set to true makes the algorithm produce intermediate output
%       (default = true). 
%
%   See also: Round, minErrRRSVD, linprog
%
%   This code is free to use for scientific purposes provided that the
%   original paper is cited.
%
%   Copyright Pauli Miettinen & Stefan Neumann, 2016


%% Check input
p = inputParser;

% Defaults
defaultThreshold = 0.5;
defaultReplicates = 1;
defaultProjection = 'sparse';
defaultSolver = 'linprog';
defaultVerbose = true;

% Valid options
validProjections = {'sparse', 'normal', 'binary', 'uniform'};
validSolvers = {'linprog', 'gurobi', 'cvx_sedumi', 'cvx_sdpt3', 'cvx_gurobi', 'cvx_mosek'};

% Build parser
addRequired(p, 'B', ...
    @(x) validateattributes(x, {'numeric'}, {'nonempty','2d','binary'}, mfilename, 'B', 1));
addRequired(p, 'k', ...
    @(x) validateattributes(x, {'numeric'}, {'scalar', 'positive', 'integer'}, mfilename, 'k', 2));
addParameter(p, 'threshold', defaultThreshold, @isscalar);
addParameter(p, 'replicates', defaultReplicates, ...
    @(x) validateattributes(x, {'numeric'}, {'scalar', 'nonnegative', 'integer'}, mfilename, 'replicates'));
addParameter(p, 'solver', defaultSolver, @ischar);
addParameter(p, 'projection', defaultProjection, @ischar);
addParameter(p, 'verbose', defaultVerbose, @isscalar);

% Parse
parse(p, B, k, varargin{:});

threshold = p.Results.threshold;
replicates = p.Results.replicates;
solver = validatestring(p.Results.solver, validSolvers, mfilename, 'solver');
projection = validatestring(p.Results.projection, validProjections, mfilename, 'projection');
verbose = logical(p.Results.verbose);

bestErr = Inf;
bestL = [];
bestR = [];


m = size(B,1);
n = size(B,2);

epsilon = 0.000001;



for run = 1:replicates
    verbprint(verbose, 'Dimension %3i, replicate %2i (of %i)\n', k, run, replicates);
    % get m points in R^d in general position
    % store these points in the rows of matrix V
    % (the points are in general position with
    %  probability 1)
    
    verbprint(verbose, '\tDoing projection\n');
    V = getpm(n, k, projection);
    L = B * V;
    
    % the matrix R with the hyperplanes
    R = zeros(k,n);
    
    verbprint(verbose, '\tFinding the RHS matrix\n');
    parfor j = 1:n
        [f, A, b, lb] = init_proj_lp(B(:,j), L, threshold, epsilon);
        [x, e, exitflag] = solveminerr(f, A, b, lb, solver);
        R(:,j) = x;
        if exitflag ~= 1
            warning('Column %i optimization raised exitflag %i', j, exitflag);
        end
        
    end
    err = sum(sum(abs(B - Round(L*R, threshold))));
    found_solution = true;
    if err < bestErr
        bestErr = err;
        bestL = L;
        bestR = R;
    end
end

verbprint(verbose, 'Finished looking for solution\n');


err = bestErr;
L = bestL;
R = bestR;
verbprint(verbose, 'Found answer with error = %d\n', err);


end

%% verbprint - print based on verbosity
function nbytes = verbprint(v, varargin)
    nbytes = 0;
    if v
        nbytes=fprintf(1, varargin{:});
    end
end

%% getpm - get the projection matrix
function V = getpm(n, d, projection)
switch projection
    case 'normal'
        V = randn(n, d);
    case 'binary'
        V = 2*double(rand(n, d) < 0.5);
        V = V - 1;
    case 'sparse'
        nonz = find(rand(n*d, 1) < 1/3);
        nonz = nonz(randperm(length(nonz)));
        cut = round(length(nonz)/2);
        [plusI, plusJ] = ind2sub([n, d], nonz(1:cut));
        [minusI, minusJ] = ind2sub([n, d], nonz((cut+1):end));
        I = [plusI; minusI];
        J = [plusJ; minusJ];
        V = [sqrt(3)*ones(length(plusI),1); -sqrt(3)*ones(length(minusI), 1)];
        V = sparse(I, J, V, n, d);
    case 'uniform'
        V = rand(n,d) - 1/2;
        V = normr(V);
    otherwise
        error('The projection method %s is not known', projection);
end
        
end

%% solvelp - solves the linear program
function [x, exitflag] = solvelp(A, b, solver)
d = size(A, 2);
switch solver
    case 'linprog'
        options = optimoptions('linprog', 'Display', 'off');
        [x, ~, exitflag] = linprog(zeros(d,1), A, full(b), [], [], [], [], [], options);
    case 'gurobi'
        [x,~,exitflag] = linprog_gurobi(zeros(d,1), A, full(b));
    case {'cvx_sedumi', 'cvx_sdpt3', 'cvx_gurobi', 'cvx_mosek'}
        alg = solver(5:end); % The name of the solver
        cvx_begin quiet
          cvx_solver(alg);
          variable x(d);
          subject to 
            A*x <= b;
        cvx_end
        if cvx_optval == 0
            exitflag = 1; % constraints are feasible
        else
            exitflag = 2; 
            x = zeros(d, 1);
        end
    otherwise
        error('Unknown solver %s', solver);
end
end

%% solveminerr - solvest the minimum-margin LP
function [x, val, exitflag] = solveminerr(f, A, b, lb, solver)
k = size(A, 2) - length(b); % A has k + m columns, get d
switch solver
    case 'linprog'
        options = optimoptions('linprog', 'Display', 'off');
        [x, val, exitflag] = linprog(f, A, full(b), [], [], lb, [], [], options);
        x = x(1:k);
  case 'gurobi'
    [x, val, exitflag] = linprog_gurobi(f, A, full(b), [], [], lb);
    x = x(1:k);
    otherwise
        error('Cannot use solver %s to solve the minimum-margin problem', solver);
end
end
