function [L, R, err] = minErrRRSVD(B, k, varargin)
%minErrRRSVD Outputs a fixed-rounding-rank approximation using SVD.
%   [L, R] = minErrRRSVD(B, k) returns the factor matrices.
%   [L, R, err] = minErrRRSVD(B, k) returns also the reconstruction error.
%   minErrRRSVD(..., 'threshold', tau) uses threshold "tau" as the 
%       rounding threshold, i.e. Round(X, tau) = B (default = 0.5)
%   minErrRRSVD(..., 'verbose', v) sets the verbosity: 0 or false
%       means no output, 1 or true means output (default = true)
%
%   See also: Round, RoundingRankSVD, minErrRRProj
%
%   This code is free to use for scientific purposes provided that the
%   original paper is cited.
%
%   Copyright Pauli Miettinen & Stefan Neumann, 2016

%% Check input
p = inputParser;
defaultThreshold = 0.5;
defaultVerbose = true;

addRequired(p, 'B', ...
    @(x) validateattributes(x, {'numeric'}, {'nonempty','2d','binary'}));
addRequired(p, 'k', ...
    @(x) validateattributes(x, {'numeric'}, {'scalar', 'positive', 'integer'}, mfilename, 'k', 2));
addParameter(p, 'threshold', defaultThreshold, @isscalar);
addParameter(p, 'verbose', defaultVerbose, @isscalar);

parse(p, B, k, varargin{:});

tau = p.Results.threshold;
verbose = logical(p.Results.verbose);


%% Do the math
if verbose
    fprintf('Computing the SVD...\n');
end
[U,S,V] = svds(B, k);

L = U;
R = S*V';

if nargout > 2
    err = nnz(B - Round(L*R, tau));
end

end

