function [ A, L, R, iter ] = quasi_n_logistic(B, r, varargin)
%QUASI_N_LOGISTIC solves logistic decomposition using Quasi-Newton
%
%  [A, L, R] = QUASI_N_LOGISTIC(B, r) finds rank-r decomposition that is
%  close to B. L and R are the factor matrices and A is logit(L*R). 
%
%  [A, L, R, iter] = QUASI_N_LOGISTIC() returns the number of iterations,
%  too.
%
%  QUASI_N_LOGISTIC(..., 'steepness', n) sets the steepness of the logistic function.
%  QUASI_N_LOGISTIC(..., 'maxiter', n) sets the number of maximum
%  iterations (default = 300).
%  QUASI_N_LOGISTIC(..., 'epsilon', e) sets the minimum improvement
%  (default = 1).
%  QUASI_N_LOGISTIC(..., 'L', L, 'R', R) sets the initial factor
%  matrices.
%
%  See also: LOGITFRO
%
%   This code is free to use for scientific purposes provided that the
%   original paper is cited.
%
%   Copyright Pauli Miettinen & Stefan Neumann, 2016

[n, m] = size(B);

defaultSteepness = 2;
defaultMaxIter = 300;
defaultEpsilon = 1;
defaultL = [];
defaultR = [];
p = inputParser;
addRequired(p, 'B', @(x) validateattributes(x, {'numeric'}, {'2d', ...
                    'binary'}, mfilename, 'B', 1));
addRequired(p, 'k', @(x) validateattributes(x, {'numeric'}, {'scalar', ...
                    'integer', 'positive'}, mfilename, 'k', 2));
addParameter(p, 'steepness', defaultSteepness, @(x) validateattributes(x, {'numeric'}, {'scalar', ...
                    'positive'}, mfilename, 'steepness'));
addParameter(p, 'maxiter', defaultMaxIter, @(x) validateattributes(x, {'numeric'}, {'scalar', ...
                    'integer', 'positive'}, mfilename, 'maxiter'));
addParameter(p, 'epsilon', defaultEpsilon, @(x) validateattributes(x, {'numeric'}, {'scalar', ...
                    'nonnegative'}, mfilename, 'epsilon'));
addParameter(p, 'L', defaultL, @(x) validateattributes(x, {'numeric'}, {'2d', ...
                    'real'}, mfilename, 'L'));
addParameter(p, 'R', defaultR, @(x) validateattributes(x, {'numeric'}, {'2d', ...
                    'real'}, mfilename, 'R'));

parse(p, B, r, varargin{:});

k       = p.Results.steepness;
maxiter = p.Results.maxiter;
epsilon = p.Results.epsilon;
if isempty(p.Results.L)
    L = rand(n,r) + sqrt(0.5 / r) - 0.5;
else
    L = p.Results.L;
end
if isempty(p.Results.R)
    R = rand(r,m) + sqrt(0.5 / r) - 0.5;
else
    R = p.Results.R;
end


oldScore = logitfro(B, L, R, k);

options = optimoptions(@fminunc, 'Algorithm', 'quasi-newton', 'Display', 'off');


iter = 1;
while iter < maxiter
    % Update L
    for row = 1:n
        fun = @(x) logitfro(B(row,:), x, R, k);
        [x, ~, exitflag] = fminunc(fun, L(row, :), options);
        if exitflag == 1
            L(row, :) = x;
        end
    end
    
    % Update R
    for col=1:m
        fun = @(x) logitfro(B(:,col), L, x, k);
        [x, ~, exitflag] = fminunc(fun, R(:, col), options);
        if exitflag == 1
            R(:, col) = x;
        end
    end
    
    newScore = logitfro(B, L, R, k);
    
    if oldScore - newScore < epsilon
        break;
    end
    oldScore = newScore;
    iter = iter + 1;
end    

A = (1 + exp(-k * (L*R - 1/2) )).^(-1);
