%RUN_REAL_ERR_APJ Finds the min-err rrank-k decomposition of the APJ data
%
%   This code is free to use for scientific purposes provided that the
%   original paper is cited.
%
%   Copyright Pauli Miettinen & Stefan Neumann, 2016

resdir = '../results/real/';
dataname = 'apj';

XY = importdata('../data/real/role_mining/apj.txt');

B = sparse(XY(:,1), XY(:,2), 1);

B(B > 1) = 1; % Make sure B is binary

tau = 0.5;
replicates = 5;


K = [5 10 15];

template = struct();
template.time = nan( length(K), 1);
template.err  = nan( length(K), 1);
template.L    = cell(length(K), 1);
template.R    = cell(length(K), 1);
template.nnz  = nan( length(K), 1);

proje = template;
SVD = template;
lpca = template;
lpca.iters = nan(length(K), 1);
truncSVD = template;
asso = template;


%% Projection w/ Gurobi
for kIter=1:length(K)
    k = K(kIter);
    fprintf(['**********************\n* Projection k = %i\n**************' ...
             '********\n'], k);
    tic
    [L, R, err] = minErrRRProj(B, k, 'threshold', tau, 'solver', 'gurobi', 'replicates', replicates);
    time = toc;
    proje.err(kIter) = err;
    proje.L{kIter} = L;
    proje.R{kIter} = R;
    proje.time(kIter) = time;
    proje.nnz(kIter) = nnz(B);
    save([resdir dataname '_err_proje'], 'proje', 'K');
end

% SVD
for kIter=1:length(K)
    k = K(kIter);
    fprintf(['**********************\n* Rounded SVD k = %i\n**************' ...
             '********\n'], k);
    tic
    [L, R, err] = minErrRRSVD(B, k, 'threshold', tau);
    time = toc;
    SVD.err(kIter) = err;
    SVD.L{kIter} = L;
    SVD.R{kIter} = R;
    SVD.time(kIter) = time;
    SVD.nnz(kIter) = nnz(B);
    save([resdir dataname '_err_svd'], 'SVD', 'K');
end

%% Logistic PCA
for kIter=1:length(K)
    k = K(kIter);
    fprintf(['**********************\n* Logistic PCA k = %i\n**************' ...
             '********\n'], k);
    tic
    [L, R, iters] = lpca_fit(B, k, 'threshold', tau, 'replicates', replicates, ...
                             'numiterations', 50);
    time = toc;
    err = sum(sum(abs(B - Round(L*R, tau))));
    lpca.err(kIter) = err;
    lpca.L{kIter} = L;
    lpca.R{kIter} = R;
    lpca.time(kIter) = time;
    lpca.nnz(kIter) = nnz(B);
    lpca.iters(kIter) = iters;
    save([resdir dataname '_err_lpca'], 'lpca', 'K');
end

%% Truncated SVD
for kIter=1:length(K)
    k = K(kIter);
    fprintf(['**********************\n* Tuncated SVD k = %i\n**************' ...
             '********\n'], k);
    B = sparse(B);
    tic
    [U, S, V] = svds(B, k);
    time = toc;
    L = U*S;
    R = V';
    err = norm(B - L*R, 'fro')^2;
    truncSVD.err(kIter) = err;
    truncSVD.L{kIter} = L;
    truncSVD.R{kIter} = R;
    truncSVD.time(kIter) = time;
    truncSVD.nnz(kIter) = nnz(B);
    save([resdir dataname '_err_truncSVD'], 'truncSVD', 'K');
end

%% Asso
for kIter=1:length(K)
    k = K(kIter);
    fprintf(['**********************\n* Asso k = %i\n**************' ...
             '********\n'], k);
    B = sparse(B);
    tic
    [L, R] = sparse_asso(B, 0.05:0.05:0.95, k, 'processes', 8);
    time = toc;
    err = bnorm(B, L, R);
    asso.err(kIter) = err;
    asso.L{kIter} = L;
    asso.R{kIter} = R;
    asso.time(kIter) = time;
    asso.nnz(kIter) = nnz(B);
    save([resdir dataname '_err_asso'], 'asso', 'K');
end
