function run_synth_err_small(varargin)
%RUN_SYNTH_ERR_SMALL Runs synthetic min-err experiments on small matrices
%
% RUN_SYNTH_ERR_SMALL('feature', false) skips testing algorith or test
%   'feature'. The effect is order-dependent, so setting 'allAlgos' to
%   false first and 'proje' to true then would only run the 'proje'
%   algo. Similarly, 'allTests' shuts down all tests.
%
%   Possible tests are 'vary_n', 'vary_k', 'vary_dens', 'vary_noise', and
%   'vary_rank'. Possible algorithms are 'SVD', 'proje', 'lpca',
%   'truncSVD', 'asso', 'sgd', and 'qnewt'. The last two methods are, by
%   default, not run.
%
% RUN_SYNTH_ERR_SMALL('suffix', str) appends str to the overall result
%   file's name.
%
% RUN_SYNTH_ERR_SMALL('append', true) appends the results to existing
%   MAT-files.
%
%   This code is free to use for scientific purposes provided that the
%   original paper is cited.
%
%   Copyright Pauli Miettinen & Stefan Neumann, 2016

%% Which experiments to run - check also parameters
tests = struct(...
    'vary_n', true, ...
    'vary_k', true, ...
    'vary_dens', true, ...
    'vary_noise', true, ...
    'vary_rank', true );

algos = struct(...
    'SVD', true, ...
    'proje', true, ...
    'lpca', true, ...
    'truncSVD', true, ...
    'asso', true, ...
    'sgd', false, ...
    'qnewt', false);

if mod(nargin, 2) ~= 0
    error('All parameters must be keyword-value pairs');
end

validAlgos = {'SVD', 'proje', 'lpca', 'truncSVD', 'asso', 'sgd', 'qnewt'};
validTests = {'vary_n', 'vary_k', 'vary_dens', ...
              'vary_noise', 'vary_rank'};
universals = {'allAlgos', 'allTests', 'suffix', 'append'};
validStrings = [validAlgos, validTests, universals];

result_suffix = ''; % use this to separate the results of multiple runs
append = false; % do we append the results to existing files

for i=1:2:nargin
    cleanStr = validatestring(varargin{i}, validStrings, ...
                              mfilename);
    switch cleanStr
      case validAlgos
        algos.(cleanStr) = logical(varargin{i+1});
      case validTests
        tests.(cleanStr) = logical(varargin{i+1});
      case 'allAlgos'
        for j=1:length(validAlgos)
            algos.(validAlgos{j}) = logical(varargin{i+1});
        end
      case 'allTests'
        for j=1:length(validTests)
            tests.(validTests{j}) = logical(varargin{i+1});
        end
      case 'suffix'
        result_suffix = varargin{i+1};
      case 'append'
        append = varargin{i+1};
        % validatestring makes sure the string is one of these
    end
end

%% Parameters
N = 60:40:220;
defN = 100;
m = 100;
K = 5:5:30;
defK = 10;

%bigN = 500;
%bigM = 300;
%bigK = 60;

tau = 0.5;

Density = 0.1:0.1:0.7;
defDensity = 0.5;

Noise = 0:0.05:0.5;
defNoise = 0;
Iterations = 10;


datadir='../data/synth';
resdir='../results/synth';



%% Different N
if tests.vary_n
    k = defK;
    noise = defNoise;
    dens = defDensity;
    
    % result structs
    res = initRes(algos, Iterations, length(N));
    
    kplus = 0;
    
    for nIter=1:length(N),
        n = N(nIter);
        for iter=1:Iterations,
            fprintf('n = %i, iter = %i Gaussian, k + %i\n', n, iter, kplus);
            load(sprintf('%s/norm_%03i_%03i_%03i_%.2f_%.2f_%02i.mat', ...
                datadir, n, m, k, dens, noise, iter));
            
            filestr = sprintf('%s/per_run/%%s_n_%%s_%03i_%03i_%03i_%.2f_%.2f_%02i.mat', ...
                              resdir, n, m, k, dens, noise, iter); % Formatstring for saving the results
            Borig = Round(L*R, tau);
            res = runErrAlgos(algos, B, Borig, k + kplus, res, 'norm', iter, nIter, filestr);
 
            fprintf('n = %i, iter = %i Uniform, k + %i\n', n, iter, kplus);
            load(sprintf('%s/unif_%03i_%03i_%03i_%.2f_%.2f_%02i.mat', ...
                datadir, n, m, k, dens, noise, iter));
            Borig = Round(L*R, tau);
            res = runErrAlgos(algos, B, Borig, k + kplus, res, 'unif', iter, nIter, filestr);
            
        end
    end
    
    if ~append
        save([resdir '/err_small_vary_n_kplus_' int2str(kplus) result_suffix '.mat'], 'N');
    end
    save([resdir '/err_small_vary_n_kplus_' int2str(kplus) result_suffix '.mat'], '-struct', ...
         'res', '-append');
    
    % k + 20 test
    % re-init result structs
    res = initRes(algos, Iterations, length(N));
    kplus = 10;
    
    for nIter=1:length(N),
        n = N(nIter);
        for iter=1:Iterations,
            fprintf('n = %i, iter = %i Gaussian, k + %i\n', n, iter, kplus);
            load(sprintf('%s/norm_%03i_%03i_%03i_%.2f_%.2f_%02i.mat', ...
                datadir, n, m, k, dens, noise, iter));
            
            filestr = sprintf('%s/per_run/%%s_n_%%s_%03i_%03i_%03i_%.2f_%.2f_%02i.mat', ...
                              resdir, n, m, k, dens, noise, iter); % Formatstring for saving the results
            Borig = Round(L*R, tau);
            res = runErrAlgos(algos, B, Borig, k+kplus, res, 'norm', iter, nIter, filestr);
 
            fprintf('n = %i, iter = %i Uniform, k + %i\n', n, iter, kplus);
            load(sprintf('%s/unif_%03i_%03i_%03i_%.2f_%.2f_%02i.mat', ...
                datadir, n, m, k, dens, noise, iter));
            Borig = Round(L*R, tau);
            res = runErrAlgos(algos, B, Borig, k+kplus, res, 'unif', iter, nIter, filestr);
            
        end
    end
    
    if ~append
        save([resdir '/err_small_vary_n_kplus_' int2str(kplus) result_suffix '.mat'], 'N');
    end
    save([resdir '/err_small_vary_n_kplus_' int2str(kplus) result_suffix '.mat'], '-struct', ...
         'res', '-append');
end % if tests.vary_n

%% Different K
if tests.vary_k
    n = defN;
    noise = defNoise;
    dens = defDensity;
    
    % result structs
    res = initRes(algos, Iterations, length(K));
    
    kplus = 0;
    
    for kIter=1:length(K),
        k = K(kIter);
        for iter=1:Iterations,
            fprintf('k = %i, iter = %i Gaussian k + %i\n', k, iter, kplus);
            load(sprintf('%s/norm_%03i_%03i_%03i_%.2f_%.2f_%02i.mat', ...
                datadir, n, m, k, dens, noise, iter), 'B', 'L', 'R');
            
            filestr = sprintf('%s/per_run/%%s_k_%%s_%03i_%03i_%03i_%.2f_%.2f_%02i.mat', ...
                              resdir, n, m, k, dens, noise, iter); % Formatstring for saving the results
            Borig = Round(L*R, tau);
            res = runErrAlgos(algos, B, Borig, k + kplus, res, 'norm', iter, kIter, ...
                               filestr);
            
            fprintf('k = %i, iter = %i Uniform, k + %i\n', k, iter, kplus);
            load(sprintf('%s/unif_%03i_%03i_%03i_%.2f_%.2f_%02i.mat', ...
                datadir, n, m, k, dens, noise, iter), 'B', 'L', 'R');
            Borig = Round(L*R, tau);
            res = runErrAlgos(algos, B, Borig, k + kplus, res, 'unif', iter, kIter, ...
                               filestr);
        end
    end
    
    if ~append
        save([resdir '/err_small_vary_k_kplus_' int2str(kplus) result_suffix '.mat'], 'K');
    end
    save([resdir '/err_small_vary_k_kplus_' int2str(kplus) result_suffix '.mat'], '-struct', 'res', ...
         '-append');
    
    % k + 20
    % Re-init result structs
    res = initRes(algos, Iterations, length(K));
    kplus = 10;
    
    for kIter=1:length(K),
        k = K(kIter);
        for iter=1:Iterations,
            fprintf('k = %i, iter = %i Gaussian k + %i\n', k, iter, kplus);
            load(sprintf('%s/norm_%03i_%03i_%03i_%.2f_%.2f_%02i.mat', ...
                datadir, n, m, k, dens, noise, iter), 'B', 'L', 'R');
            
            filestr = sprintf('%s/per_run/%%s_k_%%s_%03i_%03i_%03i_%.2f_%.2f_%02i.mat', ...
                              resdir, n, m, k, dens, noise, iter); % Formatstring for saving the results
            Borig = Round(L*R, tau);
            res = runErrAlgos(algos, B, Borig, k + kplus, res, 'norm', iter, kIter, ...
                               filestr);
            
            fprintf('k = %i, iter = %i Uniform, k + %i\n', k, iter, kplus);
            load(sprintf('%s/unif_%03i_%03i_%03i_%.2f_%.2f_%02i.mat', ...
                datadir, n, m, k, dens, noise, iter), 'B', 'L', 'R');
            Borig = Round(L*R, tau);
            res = runErrAlgos(algos, B, Borig, k + kplus, res, 'unif', iter, kIter, ...
                               filestr);
        end
    end
    if ~append
        save([resdir '/err_small_vary_k_kplus_' int2str(kplus) result_suffix '.mat'], 'K');
    end
    save([resdir '/err_small_vary_k_kplus_' int2str(kplus) result_suffix '.mat'], '-struct', 'res', '-append');
    
end % if tests.vary_k

    
    
%% Different densities
if tests.vary_dens
 

    n = defN;
    k = defK;
    noise = defNoise;
    
    % result structs
    res = initRes(algos, Iterations, length(Density));

    kplus = 0;
    
    for tDens=1:length(Density)
        dens = Density(tDens);
        for iter=1:Iterations,
            fprintf('dens = %.2f, iter = %i Uniform, k + %i\n', dens, iter, ...
                    kplus);
            load(sprintf('%s/unif_%03i_%03i_%03i_%.2f_%.2f_%02i.mat', ...
                datadir, n, m, k, dens, noise, iter));
            
            filestr = sprintf('%s/per_run/%%s_dens_%%s_%03i_%03i_%03i_%.2f_%.2f_%02i.mat', ...
                              resdir, n, m, k, dens, noise, iter); % Formatstring for saving the results
            Borig = Round(L*R, tau);
            res = runErrAlgos(algos, B, Borig, k + kplus, res, 'unif', iter, tDens, filestr);
            
            fprintf('dens = %.2f, iter = %i Gaussian\n', dens, iter);
            load(sprintf('%s/norm_%03i_%03i_%03i_%.2f_%.2f_%02i.mat', ...
                datadir, n, m, k, dens, noise, iter));
            Borig = Round(L*R, tau);
            res = runErrAlgos(algos, B, Borig, k + kplus, res, 'norm', iter, tDens, filestr);
        end
    end
    if ~append
        save([resdir '/err_small_vary_dens_kplus_' int2str(kplus) result_suffix '.mat'], 'Density');
    end
    save([resdir '/err_small_vary_dens_kplus_' int2str(kplus) result_suffix '.mat'], '-struct', 'res', ...
         '-append');

    % K + 20
    % result structs
    res = initRes(algos, Iterations, length(Density));

    kplus = 10;
    
    for tDens=1:length(Density)
        dens = Density(tDens);
        for iter=1:Iterations,
            fprintf('dens = %.2f, iter = %i Uniform, k + %i\n', dens, iter, ...
                    kplus);
            load(sprintf('%s/unif_%03i_%03i_%03i_%.2f_%.2f_%02i.mat', ...
                datadir, n, m, k, dens, noise, iter));
            
            filestr = sprintf('%s/per_run/%%s_dens_%%s_%03i_%03i_%03i_%.2f_%.2f_%02i.mat', ...
                              resdir, n, m, k, dens, noise, iter); % Formatstring for saving the results
            Borig = Round(L*R, tau);
            res = runErrAlgos(algos, B, Borig, k + kplus, res, 'unif', iter, tDens, filestr);
            
            fprintf('dens = %.2f, iter = %i Gaussian\n', dens, iter);
            load(sprintf('%s/norm_%03i_%03i_%03i_%.2f_%.2f_%02i.mat', ...
                datadir, n, m, k, dens, noise, iter));
            Borig = Round(L*R, tau);
            res = runErrAlgos(algos, B, Borig, k + kplus, res, 'norm', iter, tDens, filestr);
        end
    end
    if ~append
        save([resdir '/err_small_vary_dens_kplus_' int2str(kplus) result_suffix '.mat'], 'Density');
    end
    save([resdir '/err_small_vary_dens_kplus_' int2str(kplus) result_suffix '.mat'], '-struct', 'res', ...
         '-append');
end % if tests.vary_dens


%% Different noise leves
if tests.vary_noise


    n = defN;
    k = defK;
    dens = defDensity;
    
    % result structs
    res = initRes(algos, Iterations, length(Noise));

    kplus = 0;
    
    for iterNoise=1:length(Noise)
        noise = Noise(iterNoise);
        for iter=1:Iterations,
            fprintf('noise = %.2f, iter = %i Uniform, k + %i\n', noise, ...
                    iter, kplus);
            load(sprintf('%s/unif_%03i_%03i_%03i_%.2f_%.2f_%02i.mat', ...
                datadir, n, m, k, dens, noise, iter));
            
            filestr = sprintf('%s/per_run/%%s_noise_%%s_%03i_%03i_%03i_%.2f_%.2f_%02i.mat', ...
                              resdir, n, m, k, dens, noise, iter); % Formatstring for saving the results
            
            res = runErrAlgos(algos, B, Borig, k + kplus, res, 'unif', iter, iterNoise, filestr);
            
            fprintf('noise = %.2f, iter = %i Gaussian, k + %i\n', noise, ...
                    iter, kplus);
            load(sprintf('%s/norm_%03i_%03i_%03i_%.2f_%.2f_%02i.mat', ...
                datadir, n, m, k, dens, noise, iter));
            
            res = runErrAlgos(algos, B, Borig, k + kplus, res, 'norm', iter, iterNoise, filestr);
        end
    end
    if ~append
        save([resdir '/err_small_vary_noise_kplus_' int2str(kplus) result_suffix '.mat'], 'Noise');
    end
    save([resdir '/err_small_vary_noise_kplus_' int2str(kplus) result_suffix '.mat'], '-struct', 'res', ...
         '-append');

    % result structs
    res = initRes(algos, Iterations, length(Noise));

    kplus = 10;
    
    for iterNoise=1:length(Noise)
        noise = Noise(iterNoise);
        for iter=1:Iterations,
            fprintf('noise = %.2f, iter = %i Uniform, k + %i\n', noise, ...
                    iter, kplus);
            load(sprintf('%s/unif_%03i_%03i_%03i_%.2f_%.2f_%02i.mat', ...
                datadir, n, m, k, dens, noise, iter));
            
            filestr = sprintf('%s/per_run/%%s_noise_%%s_%03i_%03i_%03i_%.2f_%.2f_%02i.mat', ...
                              resdir, n, m, k, dens, noise, iter); % Formatstring for saving the results
            
            res = runErrAlgos(algos, B, Borig, k + kplus, res, 'unif', iter, iterNoise, filestr);
            
            fprintf('noise = %.2f, iter = %i Gaussian, k + %i\n', noise, ...
                    iter, kplus);
            load(sprintf('%s/norm_%03i_%03i_%03i_%.2f_%.2f_%02i.mat', ...
                datadir, n, m, k, dens, noise, iter));
            
            res = runErrAlgos(algos, B, Borig, k + kplus, res, 'norm', iter, iterNoise, filestr);
        end
    end
    if ~append
        save([resdir '/err_small_vary_noise_kplus_' int2str(kplus) result_suffix '.mat'], 'Noise');
    end
    save([resdir '/err_small_vary_noise_kplus_' int2str(kplus) result_suffix '.mat'], '-struct', 'res', ...
         '-append');
    
end % if tests.vary_noise

%% Different decomposition ranks
if tests.vary_rank


    n = defN;
    k = defK;
    dens = defDensity;
    noise = 0.25;
    
    Rank = 5:5:50;
    
    % result structs
    res = initRes(algos, Iterations, length(Rank));
    
    for iterRank=1:length(Rank)
        rank = Rank(iterRank);
        for iter=1:Iterations,
            fprintf('rank = %i, iter = %i Uniform\n', rank, ...
                    iter);
            load(sprintf('%s/unif_%03i_%03i_%03i_%.2f_%.2f_%02i.mat', ...
                datadir, n, m, k, dens, noise, iter));
            
            filestr = sprintf('%s/per_run/%%s_noise_%%s_%03i_%03i_%03i_%.2f_%.2f_%03i_%02i.mat', ...
                              resdir, n, m, k, dens, noise, rank, iter); % Formatstring for saving the results
            %Borig = Round(L*R, tau);
            res = runErrAlgos(algos, B, Borig, rank, res, 'unif', iter, iterRank, filestr);
            
            fprintf('rank = %i, iter = %i Gaussian\n', rank, ...
                    iter);
            load(sprintf('%s/norm_%03i_%03i_%03i_%.2f_%.2f_%02i.mat', ...
                datadir, n, m, k, dens, noise, iter));
            %Borig = Round(L*R, tau);
            res = runErrAlgos(algos, B, Borig, rank, res, 'norm', iter, iterRank, filestr);
        end
    end
    if ~append
        save([resdir '/err_small_vary_rank' result_suffix '.mat'], 'Rank');
    end
    save([resdir '/err_small_vary_rank' result_suffix '.mat'], '-struct', 'res', ...
         '-append');
end % if test.vary_rank

end % function run_synth_err_big
    
function res = initRes(algos, rows, cols)
template1 = struct(...
    'err',     nan(rows, cols), ...
    'errOrig', nan(rows, cols), ...
    'nnz',     nan(rows, cols), ...
    'nnzOrig', nan(rows, cols), ...
    'LL',      nan(rows, cols), ...
    'time',    nan(rows, cols) );
template = struct();
template.norm = template1;
template.unif = template1;

res = struct();
if algos.SVD
    res.SVD = template;
end
if algos.proje
    res.proje = template;
end
if algos.lpca
    res.lpca = template;
    % LPCA also stores the iterations
    res.lpca.norm.iters = nan(rows, cols);
    res.lpca.unif.iters = nan(rows, cols);
end
if algos.truncSVD
    res.truncSVD = template;
end
if algos.asso
    res.asso = template;
end
if algos.sgd
    res.sgd = template;
    % SGD also stores the iterations
    res.sgd.norm.iters = nan(rows, cols);
    res.sgd.unif.iters = nan(rows, cols);    
end
if algos.qnewt
    res.qnewt = template;
    % Quasi-Newton also stores the iterations
    res.qnewt.norm.iters = nan(rows, cols);
    res.qnewt.unif.iters = nan(rows, cols);    
end
end % Function initRes

function res = runErrAlgos(algos, B, Borig, k, res, type, row, col, filestr, varargin)
    defaultThreshold  = 0.5;
    defaultReplicates = 5;
    defaultVerbose    = false;
    p = inputParser;
    addRequired(p, 'algos', @(x) ...
                validateattributes(x, {'struct'}, {'nonempty'}, 'runErrAlgos', ...
                                   'algos', 1));
    addRequired(p, 'B', @(x) validateattributes(x, {'numeric'}, {'2d', ...
                        'binary'}, 'runErrAlgos', 'B', 2));
    addRequired(p, 'Borig', @(x) validateattributes(x, {'numeric'}, {'2d', ...
                        'binary'}, 'runErrAlgos', 'Borig', 3));
    addRequired(p, 'k', @(x) validateattributes(x, {'numeric'}, {'scalar', ...
                   'integer', 'positive'}, 'runErrAlgos', 'k', 4));
    addRequired(p, 'res', ...
                @(x) validateattributes(x, {'struct'}, {'nonempty'}, ...
                                        'runErrAlgos', 'res', 5));
    addRequired(p, 'type', ...
                @(x) any(validatestring(x, {'norm', 'unif'}, 'runErrAlgos', ...
                                    'type', 6)));
    addRequired(p, 'row', ...
                @(x) validateattributes(x, {'numeric'}, {'scalar', 'integer', ...
                   'positive'}, 'runErrAlgos', 'row', 7));
    addRequired(p, 'col', ...
                @(x) validateattributes(x, {'numeric'}, {'scalar', 'integer', ...
                   'positive'}, 'runErrAlgos', 'col', 8));
    addRequired(p, 'filestr', @ischar);
    addOptional(p, 'threshold', defaultThreshold, ...
                @(x) validateattributes(x, {'numeric'}, {'scalar', 'real'}, ...
                                        'runErrAlgos', 'threshold'));
    addOptional(p, 'replicates', defaultReplicates, ...
                @(x) validateattributes(x, {'numeric'}, {'scalar', 'integer' ...
                   'positive'}, 'runErrAlgos', 'replicates'));
    addOptional(p, 'verbose', defaultVerbose, ...
                @(x) validateattributes(x, {'numeric', 'logical'}, {'scalar'}, ...
                                        'runErrAlgos', 'verbose'));
    parse(p, algos, B, Borig, k, res, type, row, col, filestr, varargin{:});
    
    tau = p.Results.threshold;
    replicates = p.Results.replicates;
    verbose = p.Results.verbose;

    
    if algos.SVD
        fprintf('\tSVD\n');
        tic
        [L, R, err] = minErrRRSVD(B, k, 'verbose', verbose, 'threshold', tau);
        t = toc;
        X = Round(L*R, tau);
        res.SVD.(type).err(    row, col) = err;
        res.SVD.(type).errOrig(row, col) = sum(sum(abs(Borig - X)));
        res.SVD.(type).LL(     row, col) = loglikelihood(B, X); 
        res.SVD.(type).time(   row, col) = t;
        res.SVD.(type).nnz(    row, col) = nnz(B);
        res.SVD.(type).nnzOrig(row, col) = nnz(Borig);
        save(sprintf(filestr, 'svd_err', type), 'err', 'L', 'R');
    end
    
    if algos.proje
        fprintf('\tprojection\n');
        tic
        [L, R, err] = minErrRRProj(B, k, 'verbose', verbose, 'threshold', ...
                                  tau, 'replicates', replicates, 'solver', ...
                                  'gurobi');
        t = toc;
        X = Round(L*R, tau);
        res.proje.(type).err(    row, col) = err;
        res.proje.(type).errOrig(row, col) = sum(sum(abs(Borig - X)));
        res.proje.(type).LL(     row, col) = loglikelihood(B, X); 
        res.proje.(type).time(   row, col) = t;
        res.proje.(type).nnz(    row, col) = nnz(B);
        res.proje.(type).nnzOrig(row, col) = nnz(Borig); 
        save(sprintf(filestr, 'proje_err', type), 'err', 'L', 'R');
    end
    
    if algos.lpca
        fprintf('\tLogistic PCA\n');
        tic
        [L, R, iters] = lpca_fit(B, k, 'verbose', verbose, 'threshold', ...
                                         tau, 'replicates', replicates, ...
                                         'numiterations', 50);
        t = toc;
        X = Round(L*R, tau);
        err = sum(sum(abs(B - X)));
        res.lpca.(type).err(    row, col) = err;
        res.lpca.(type).errOrig(row, col) = sum(sum(abs(Borig - X)));
        res.lpca.(type).LL(     row, col) = loglikelihood(B, X); 
        res.lpca.(type).time(   row, col) = t;
        res.lpca.(type).nnz(    row, col) = nnz(B);
        res.lpca.(type).nnzOrig(row, col) = nnz(Borig);  
        res.lpca.(type).iters(  row, col) = iters;
        save(sprintf(filestr, 'lpca_err', type), 'err', 'L', 'R');
    end            
    
    if algos.truncSVD
        fprintf('\tTruncated SVD\n');
        B = sparse(B);
        tic
        [U, S, V] = svds(B, k);
        t = toc;
        X = U*S*V';
        err = norm(B - X, 'fro')^2;
        res.truncSVD.(type).err(    row, col) = err;
        res.truncSVD.(type).errOrig(row, col) = norm(Borig - X, 'fro')^2;
        res.truncSVD.(type).LL(     row, col) = loglikelihood(B, X); 
        res.truncSVD.(type).time(   row, col) = t;
        res.truncSVD.(type).nnz(    row, col) = nnz(B);
        res.truncSVD.(type).nnzOrig(row, col) = nnz(Borig);  
        save(sprintf(filestr, 'truncSVD_err', type), 'err', 'U', 'S', 'V');       
    end
    
    if algos.asso
        fprintf('\tAsso\n');
        B = sparse(B);
        tic
        [L, R] = sparse_asso(B, 0.05:0.05:0.95, k, 'processes', 8, 'verbosity', 0);
        t = toc;
        X = bprod(L, R);
        err = sum(sum(abs(B - X)));
        res.asso.(type).err(    row, col) = err;
        res.asso.(type).errOrig(row, col) = sum(sum(abs(Borig - X)));
        res.asso.(type).LL(     row, col) = loglikelihood(B, X); 
        res.asso.(type).time(   row, col) = t;
        res.asso.(type).nnz(    row, col) = nnz(B);
        res.asso.(type).nnzOrig(row, col) = nnz(Borig);  
        save(sprintf(filestr, 'asso_err', type), 'err', 'L', 'R');       
    end

    if algos.sgd
        fprintf('\tSGD\n');
        tic
        [~, L, R, iters] = StochasticGradientDescent(B, k);
        t = toc;
        X = Round(L*R, tau);
        err = sum(sum(abs(B - X)));
        res.sgd.(type).err(    row, col) = err;
        res.sgd.(type).errOrig(row, col) = sum(sum(abs(Borig - X)));
        res.sgd.(type).LL(     row, col) = loglikelihood(B, X); 
        res.sgd.(type).time(   row, col) = t;
        res.sgd.(type).nnz(    row, col) = nnz(B);
        res.sgd.(type).nnzOrig(row, col) = nnz(Borig);  
        res.sgd.(type).iters(  row, col) = iters;
        save(sprintf(filestr, 'sgd_err', type), 'err', 'L', 'R');
    end
    
    if algos.qnewt
        fprintf('\tQuasi-Newton\n');
        tic
        [~, L, R, iters] = quasi_n_logistic(B, k);
        t = toc;
        X = Round(L*R, tau);
        err = sum(sum(abs(B - X)));
        res.qnewt.(type).err(    row, col) = err;
        res.qnewt.(type).errOrig(row, col) = sum(sum(abs(Borig - X)));
        res.qnewt.(type).LL(     row, col) = loglikelihood(B, X); 
        res.qnewt.(type).time(   row, col) = t;
        res.qnewt.(type).nnz(    row, col) = nnz(B);
        res.qnewt.(type).nnzOrig(row, col) = nnz(Borig);  
        res.qnewt.(type).iters(  row, col) = iters;
        save(sprintf(filestr, 'qnewt_err', type), 'err', 'L', 'R');
    end        
end
