function run_synth_nested(varargin)
%RUN_SYNTH_NESTED runs synthetic experiments on nested matrices
%
%   RUN_SYNTH_NESTED('feature', false) skips testing algorith or test
%   'feature'. The effect is order-dependent, so setting 'allAlgos' to
%   false first and 'mt' to true after would only run the 'mt'
%   algoright. Similarly, 'allTests', false shuts down all tests.
%
%   The possible algorithms are 'SVD' (for R-SVD in the paper), 'mt' (for
%   MT), 'greedy' (for NNRR1), 'mt_greedy' (for MT+NNRR1), and 'gradient'
%   and 'mt_gradient' for an unpublished algorithm. 'allAlgos' will turn
%   all algorithms on or off.
%
%   The possible experiments are 'vary_dens' and 'vary_noise' and
%   'allTests' will turn all experiments on or off. 
%
%   RUN_SYNTH_NESTED('suffix', str) appends str to the overall result
%   file's name.
%
%   RUN_SYNTH_NESTED('append', true) will append the current results to
%   existing MAT files.
%
%   This code is free to use for scientific purposes provided that the
%   original paper is cited.
%
%   Copyright Pauli Miettinen & Stefan Neumann, 2016

%% Which experiments to run - check also parameters
    tests = struct(...
        'vary_dens', true, ...
        'vary_noise', true );

    algos = struct(...
        'mt', true, ...
        'SVD', true, ...
        'greedy', true, ...
        'gradient', false, ...
        'mt_greedy', true, ...
        'mt_gradient', false );

    if mod(nargin, 2) ~= 0
        error('All parameters must be keyword-value pairs');
    end

    validAlgos = {'SVD', 'mt', 'greedy', 'gradient', 'mt_greedy', 'mt_gradient'};
    validTests = {'vary_dens', 'vary_noise'};
    universals = {'allAlgos', 'allTests', 'suffix', 'append'};
    validStrings = [validAlgos, validTests, universals];

    result_suffix = ''; % use this to separate the results of multiple runs
    append = false; % do we append the results to existing files

    for i=1:2:nargin
        cleanStr = validatestring(varargin{i}, validStrings, ...
                                  mfilename);
        switch cleanStr
          case validAlgos
            algos.(cleanStr) = logical(varargin{i+1});
          case validTests
            tests.(cleanStr) = logical(varargin{i+1});
          case 'allAlgos'
            for j=1:length(validAlgos)
                algos.(validAlgos{j}) = logical(varargin{i+1});
            end
          case 'allTests'
            for j=1:length(validTests)
                tests.(validTests{j}) = logical(varargin{i+1});
            end
          case 'suffix'
            result_suffix = varargin{i+1};
          case 'append'
            append = varargin{i+1};
            % validatestring makes sure the string is one of these
        end
    end

    %% Parameters
    n = 200;
    m = 300;

    Density = 0.1:0.1:0.7;
    Noise   = 0.05:0.05:0.5;

    defDensity = 0.5;
    defNoise = 0.15;

    Iterations = 10;

    datadir = '../data/synth/';
    resdir='../results/synth';

    %% Vary density
    if tests.vary_dens
        noise = defNoise;
        
        % result structs
        res = initRes(algos, Iterations, length(Density));
        
        for i=1:length(Density)
            dens = Density(i);
            for iter=1:Iterations
                fprintf('dens = %.2f, iter = %i\n', dens, iter);
                load(sprintf('%s/nested_%03i_%03i_%.2f_%.2f_%02i.mat', ...
                             datadir, n, m, dens, noise, iter));
                filestr = sprintf(['%s/per_run/' ...
                                   'nested_%%s_dens_%03i_%03i_%.2f_%.2f_%02i.mat'], ...
                                  resdir, n, m, dens, noise, iter); 
                res = runAlgos(algos, B, Borig, res, iter, i, filestr);
            end
        end
        if ~append
            save([resdir '/nested_vary_dens' result_suffix '.mat'], ...
                 'Density');
        end
        save([resdir '/nested_vary_dens' result_suffix '.mat'], '-struct', ...
             'res', '-append');
    end % test.vary_dens

    %% Vary noise
    if tests.vary_noise
        dens = defDensity;
        
        % result structs
        res = initRes(algos, Iterations, length(Noise));
        
        for i=1:length(Noise)
            noise = Noise(i);
            for iter=1:Iterations
                fprintf('noise = %.2f, iter = %i\n', noise, iter);
                load(sprintf('%s/nested_%03i_%03i_%.2f_%.2f_%02i.mat', ...
                             datadir, n, m, dens, noise, iter));
                filestr = sprintf(['%s/per_run/' ...
                                   'nested_%%s_noise_%03i_%03i_%.2f_%.2f_%02i.mat'], ...
                                  resdir, n, m, dens, noise, iter); 
                res = runAlgos(algos, B, Borig, res, iter, i, filestr);
            end
        end
        if ~append
            save([resdir '/nested_vary_noise' result_suffix '.mat'], ...
                 'Noise');
        end
        save([resdir '/nested_vary_noise' result_suffix '.mat'], '-struct', ...
             'res', '-append');
    end % test.vary_noise

end % function run_synth_nested
    
    function res = initRes(algos, rows, cols)
        template = struct(...
            'err',     nan(rows, cols), ...
            'errOrig', nan(rows, cols), ...
            'nnz',     nan(rows, cols), ...
            'nnzOrig', nan(rows, cols), ...
            'time',    nan(rows, cols) );
        res = struct();
        
        if algos.mt
            res.mt = template;
        end
        if algos.SVD
            res.SVD = template;
        end
        if algos.greedy
            res.greedy = template;
            res.greedy.iters = nan(rows, cols);
        end
        if algos.gradient
            res.gradient = template;
        end
        if algos.mt_greedy
            res.mt_greedy = template;
            res.mt_greedy.iters = nan(rows, cols);
        end
        if algos.mt_gradient
            res.mt_gradient = template;
        end
    end % Function initRes
    
    function res = runAlgos(algos, B, Borig, res, row, col, filestr, ...
                            varargin)
        defaultThreshold  = 0.5;
        defaultReplicates = 5;
        defaultVerbose    = false;
        p = inputParser;
        addRequired(p, 'algos', @(x) ...
                    validateattributes(x, {'struct'}, {'nonempty'}, 'runErrAlgos', ...
                                       'algos', 1));
        addRequired(p, 'B', @(x) validateattributes(x, {'numeric'}, {'2d', ...
                            'binary'}, 'runErrAlgos', 'B', 2));
        addRequired(p, 'Borig', @(x) validateattributes(x, {'numeric'}, {'2d', ...
                            'binary'}, 'runErrAlgos', 'Borig', 3));
        addRequired(p, 'res', ...
                    @(x) validateattributes(x, {'struct'}, {'nonempty'}, ...
                                            'runErrAlgos', 'res', 4));
        addRequired(p, 'row', ...
                    @(x) validateattributes(x, {'numeric'}, {'scalar', 'integer', ...
                            'positive'}, 'runErrAlgos', 'row', 5));
        addRequired(p, 'col', ...
                    @(x) validateattributes(x, {'numeric'}, {'scalar', 'integer', ...
                            'positive'}, 'runErrAlgos', 'col', 6));
        addRequired(p, 'filestr', @ischar);
        addOptional(p, 'threshold', defaultThreshold, ...
                    @(x) validateattributes(x, {'numeric'}, {'scalar', 'real'}, ...
                                            'runErrAlgos', 'threshold'));
        addOptional(p, 'replicates', defaultReplicates, ...
                    @(x) validateattributes(x, {'numeric'}, {'scalar', 'integer' ...
                            'positive'}, 'runErrAlgos', 'replicates'));
        addOptional(p, 'verbose', defaultVerbose, ...
                    @(x) validateattributes(x, {'numeric', 'logical'}, {'scalar'}, ...
                                            'runErrAlgos', 'verbose'));
        parse(p, algos, B, Borig, res, row, col, filestr, varargin{:});
        
        tau = p.Results.threshold;
        replicates = p.Results.replicates;
        verbose = p.Results.verbose;

        %B = sparse(B);
        
        if algos.mt
            fprintf('\tMannila--Terzi\n');
            tic
            [err, X] = Nestedness(B);
            t = toc;
            X = double(X);
            if ~IsNested(X)
                warning('Mannila--Terzi didn''t return a nested matrix');
                X = nan(size(B));
            end
            res.mt.err(    row, col) = sum(sum(abs(B - X)));
            res.mt.errOrig(row, col) = sum(sum(abs(Borig - X)));
            res.mt.time(   row, col) = t;
            res.mt.nnz(    row, col) = nnz(B);
            res.mt.nnzOrig(row, col) = nnz(Borig);
            save(sprintf(filestr, 'mt'), 'err', 'X');
            mtX = X; % for future use
        end
        
        if algos.SVD
            fprintf('\tSVD\n');
            tic
            [u, s, v] = svds(B, 1);
            t = toc;
            L = u*s;
            R = v';
            X = Round(L*R, tau);
            res.SVD.err(    row, col) = sum(sum(abs(B - X)));
            res.SVD.errOrig(row, col) = sum(sum(abs(Borig - X)));
            res.SVD.time(   row, col) = t;
            res.SVD.nnz(    row, col) = nnz(B);
            res.SVD.nnzOrig(row, col) = nnz(Borig);
            save(sprintf(filestr, 'svd'), 'err', 'L', 'R');
        end
        
        if algos.greedy
            fprintf('\tGreedy\n');
            tic
            [err, L, R, iters] = iterNestedNNRank1(B);
            t = toc;
            X = Round(L*R, tau);
            res.greedy.err(    row, col) = sum(sum(abs(B - X)));
            res.greedy.errOrig(row, col) = sum(sum(abs(Borig - X)));
            res.greedy.time(   row, col) = t;
            res.greedy.iters(  row, col) = iters;
            res.greedy.nnz(    row, col) = nnz(B);
            res.greedy.nnzOrig(row, col) = nnz(Borig);
            save(sprintf(filestr, 'greedy'), 'err', 'L', 'R', 'iters');
        end
        
        if algos.gradient
            fprintf('\tGradient\n');
            tic
            [~, L, R] = ClosestNonnegativeRoundingRankOne(B);
            t = toc;
            X = Round(L*R, tau);
            res.gradient.err(    row, col) = sum(sum(abs(B - X)));
            res.gradient.errOrig(row, col) = sum(sum(abs(Borig - X)));
            res.gradient.time(   row, col) = t;
            res.gradient.iters(  row, col) = iters;
            res.gradient.nnz(    row, col) = nnz(B);
            res.gradient.nnzOrig(row, col) = nnz(Borig);
            save(sprintf(filestr, 'gradient'), 'err', 'L', 'R');
        end
        
        if algos.mt_greedy
            fprintf('\tMT-greedy\n');
            if exist('mtX', 'var') ~= 1
                [~, mtX] = Nestedness(B);
                mtX = double(mtX);
            end
            if ~IsNested(mtX)
                warning('Mannila--Terzi did not return a nested matrix');
                X = nan(size(B));
                t = nan;
                L = []; R = []; mtL = []; mtR = [];
            else
                tic
                [~, mtL, mtR] = NonnegativeRoundingRankOne(mtX);
                [err, L, R, iters] = iterNestedNNRank1(B, 0.5, 300, mtR, mtL);
                t = toc;
                X = Round(L*R, tau);
            end
            res.mt_greedy.err(    row, col) = sum(sum(abs(B - X)));
            res.mt_greedy.errOrig(row, col) = sum(sum(abs(Borig - X)));
            res.mt_greedy.time(   row, col) = t;
            res.mt_greedy.iters(  row, col) = iters;
            res.mt_greedy.nnz(    row, col) = nnz(B);
            res.mt_greedy.nnzOrig(row, col) = nnz(Borig);
            save(sprintf(filestr, 'mt_greedy'), 'err', 'L', 'R', 'iters', ...
                 'mtL', 'mtR');
        end
        
        if algos.mt_gradient
            fprintf('\tMT-gradient\n');
            if exist('mtX', 'var') ~= 1
                [~, mtX] = Nestedness(B);
                mtX = double(mtX);
            end
            if ~IsNested(mtX)
                warning('Mannila--Terzi didn''t return a nested matrix');
                X = nan(size(B));
                L = [];
                R = [];
                mtL = [];
                mtR = [];
                t = nan;
            else
                tic
                [~, mtL, mtR] = NonnegativeRoundingRankOne(mtX);
                [~, L, R] = ClosestNonnegativeRoundingRankOne(B, 8, ...
                                                          0.00001, 1000, 0.00001, mtR, mtL);
                t = toc;
                X = Round(L*R, tau);
            end
            res.mt_gradient.err(    row, col) = sum(sum(abs(B - X)));
            res.mt_gradient.errOrig(row, col) = sum(sum(abs(Borig - X)));
            res.mt_gradient.time(   row, col) = t;
            res.mt_gradient.nnz(    row, col) = nnz(B);
            res.mt_gradient.nnzOrig(row, col) = nnz(Borig);
            save(sprintf(filestr, 'mt_gradient'), 'err', 'L', 'R',  ...
                 'mtL', 'mtR');
        end
    end % function runAlgs
