function run_synth_rank_big(varargin)
%RUN_SYNTH_RANK_BIG Runs synthetic rank experiments on big matrices
%
% RUN_SYNTH_RANK_BIG('feature', false) skips testing algorith or test
%    'feature'. The effect is order-dependent, so setting 'allAlgos' to
%   false first and 'proje' to true then would only run the 'proje'
%   algo. Similarly, 'allTests' shuts down all tests.
%
%   Possible tests are 'vary_n', 'vary_k', 'vary_proj', 'vary_dens', and
%   'vary_noise'. Possible algorithms are 'SVD', 'nuclear', 'proje',
%   'shay', 'lpca', and 'LB'.
%
% RUN_SYNTH_RANK_BIG('suffix', str) appends str to the overall result
%   file's name.
%
%   This code is free to use for scientific purposes provided that the
%   original paper is cited.
%
%   Copyright Pauli Miettinen & Stefan Neumann, 2016

%% Which experiments to run - check also parameters
tests = struct(...
    'vary_n', true, ...
    'vary_k', true, ...
    'vary_proj', true, ...
    'vary_dens', true, ...
    'vary_noise', true ...
    );

% Shay's and Nuclear can't scale
algos = struct(...
    'SVD', true, ...
    'nuclear', false, ...
    'proje', true, ...
    'shay', false, ...
    'lpca', true, ...
    'LB', true);

if mod(nargin, 2) ~= 0
    error('All parameters must be keyword-value pairs');
end

validAlgos = {'SVD', 'nuclear', 'proje', 'shay', 'lpca', 'LB'};
validTests = {'vary_n', 'vary_k', 'vary_proj', 'vary_dens', ...
              'vary_noise'};
universals = {'allAlgos', 'allTests', 'suffix'};
validStrings = [validAlgos, validTests, universals];

result_suffix = ''; % use this to separate the results of multiple runs

for i=1:2:nargin
    cleanStr = validatestring(varargin{i}, validStrings, ...
                              'run_synth_rank_small');
    switch cleanStr
      case validAlgos
        algos.(cleanStr) = logical(varargin{i+1});
      case validTests
        tests.(cleanStr) = logical(varargin{i+1});
      case 'allAlgos'
        for j=1:length(validAlgos)
            algos.(validAlgos{j}) = logical(varargin{i+1});
        end
      case 'allTests'
        for j=1:length(validTests)
            tests.(validTests{j}) = logical(varargin{i+1});
        end
      case 'suffix'
        result_suffix = varargin{i+1};
        % validatestring makes sure the string is one of these
    end
end

%% Parameters
N = 400:50:600;
defN = 500;
m = 300;
K = 40:20:100;
defK = 60;

%bigN = 500;
%bigM = 300;
%bigK = 60;

tau = 0.5;

Density = 0.1:0.1:0.7;
defDensity = 0.5;

Noise = 0:0.05:0.5;
defNoise = 0;
Iterations = 5;

datadir='../data/synth';
resdir='../results/synth';



%% Different N
if tests.vary_n
    k = defK;
    noise = defNoise;
    dens = defDensity;
    
    % result structs
    res = initRes(algos, Iterations, length(N));
    
    algosN = algos;
    algosN.shay = false; % Shay's can't run this one!
    
    for nIter=1:length(N),
        n = N(nIter);
        for iter=1:Iterations,
            fprintf('n = %i, iter = %i Gaussian\n', n, iter);
            load(sprintf('%s/norm_%03i_%03i_%03i_%.2f_%.2f_%02i.mat', ...
                datadir, n, m, k, dens, noise, iter));
            
            filestr = sprintf('%s/per_run/%%s_n_%%s_%03i_%03i_%03i_%.2f_%.2f_%02i.mat', ...
                              resdir, n, m, k, dens, noise, iter); % Formatstring for saving the results
            
            res = runRankAlgos(algosN, B, res, 'norm', iter, nIter, filestr);
 
            fprintf('n = %i, iter = %i Uniform\n', n, iter);
            load(sprintf('%s/unif_%03i_%03i_%03i_%.2f_%.2f_%02i.mat', ...
                datadir, n, m, k, dens, noise, iter));

            res = runRankAlgos(algosN, B, res, 'unif', iter, nIter, filestr);
            
        end
    end
    
    save([resdir '/big_vary_n_all' result_suffix '.mat'], 'N');
    save([resdir '/big_vary_n_all' result_suffix '.mat'], '-struct', 'res', '-append');
end % if tests.vary_n

%% Different K
if tests.vary_k
    n = defN;
    noise = defNoise;
    dens = defDensity;
    
    % result structs
    res = initRes(algos, Iterations, length(K));
    
    for kIter=1:length(K),
        k = K(kIter);
        for iter=1:Iterations,
            fprintf('k = %i, iter = %i Gaussian\n', k, iter);
            load(sprintf('%s/norm_%03i_%03i_%03i_%.2f_%.2f_%02i.mat', ...
                datadir, n, m, k, dens, noise, iter), 'B', 'L', 'R');
            
            filestr = sprintf('%s/per_run/%%s_k_%%s_%03i_%03i_%03i_%.2f_%.2f_%02i.mat', ...
                              resdir, n, m, k, dens, noise, iter); % Formatstring for saving the results
            
            res = runRankAlgos(algosN, B, res, 'norm', iter, kIter, ...
                               filestr);
            
            fprintf('k = %i, iter = %i Uniform\n', n, iter);
            load(sprintf('%s/unif_%03i_%03i_%03i_%.2f_%.2f_%02i.mat', ...
                datadir, n, m, k, dens, noise, iter), 'B', 'L', 'R');
           
            res = runRankAlgos(algos, B, res, 'unif', iter, kIter, ...
                               filestr);
        end
    end
    save([resdir '/big_vary_k_all' result_suffix '.mat'], 'K');
    save([resdir '/big_vary_k_all' result_suffix '.mat'], '-struct', 'res', '-append');
end % if tests.vary_k


%% Different projections
if tests.vary_proj && algos.proje
    k = defK;
    noise = defNoise;
    dens = defDensity;
    
    projections = {'sparse', 'normal', 'binary', 'uniform'};
    proje = struct();
    for pr=projections,
        proje.(pr{:}) = ...
            struct('rank', nan(Iterations, length(N)), ...
            'time', nan(Iterations, length(N)) );
    end
    
    for nIter=1:length(N),
        n = N(nIter);
        for iter=1:Iterations,
            fprintf('n = %i, iter = %i Uniform\n', n, iter);
            load(sprintf('%s/unif_%03i_%03i_%03i_%.2f_%.2f_%02i.mat', ...
                datadir, n, m, k, dens, noise, iter));
            
            for prId = 1:length(projections)
                pr = projections{prId};
                fprintf('\tProjection %s\n', pr);
                tic
                [r, X] = RoundingRankProj(B, 'verbose', false, 'threshold', tau, 'replicates', 5, ...
                    'projection', pr, 'solver', 'gurobi');
                t = toc;
                proje.(pr).rank(iter, nIter) = r;
                proje.(pr).time(iter, nIter) = t;
                save(sprintf('%s/per_run/proje_%s_unif_%03i_%03i_%03i_%.2f_%.2f_%02i.mat', ...
                    resdir, pr, n, m, k, dens, noise, iter), 'r', 'X');
            end
        end
    end
    save([resdir '/big_vary_projections_and_n_all.mat'], 'proje', 'projections', 'N');
end % if tests.vary_proj && algos.proje
    
    
%% Different densities
if tests.vary_dens
    % result structs
    res = initRes(algos, Iterations, length(Density));


    n = defN;
    k = defK;
    noise = defNoise;
    
    for tDens=1:length(Density)
        dens = Density(tDens);
        for iter=1:Iterations,
            fprintf('dens = %.2f, iter = %i Uniform\n', dens, iter);
            load(sprintf('%s/unif_%03i_%03i_%03i_%.2f_%.2f_%02i.mat', ...
                datadir, n, m, k, dens, noise, iter));
            
            filestr = sprintf('%s/per_run/%%s_dens_%%s_%03i_%03i_%03i_%.2f_%.2f_%02i.mat', ...
                              resdir, n, m, k, dens, noise, iter); % Formatstring for saving the results
            
            res = runRankAlgos(algos, B, res, 'unif', iter, tDens, filestr);
            
            fprintf('dens = %.2f, iter = %i Gaussian\n', dens, iter);
            load(sprintf('%s/norm_%03i_%03i_%03i_%.2f_%.2f_%02i.mat', ...
                datadir, n, m, k, dens, noise, iter));
            
            res = runRankAlgos(algos, B, res, 'norm', iter, tDens, filestr);
        end
    end
    save([resdir '/big_vary_dens_all' result_suffix '.mat'], 'Density');
    save([resdir '/big_vary_dens_all' result_suffix '.mat'], '-struct', 'res', ...
         '-append');

end % if tests.vary_dens

%% Different noise leves
if tests.vary_noise
    % result structs
    res = initRes(algos, Iterations, length(Noise));


    n = defN;
    k = defK;
    dens = defDensity;
    
    for iterNoise=1:length(Noise)
        noise = Noise(iterNoise);
        for iter=1:Iterations,
            fprintf('noise = %.2f, iter = %i Uniform\n', noise, iter);
            load(sprintf('%s/unif_%03i_%03i_%03i_%.2f_%.2f_%02i.mat', ...
                datadir, n, m, k, dens, noise, iter));
            
            filestr = sprintf('%s/per_run/%%s_noise_%%s_%03i_%03i_%03i_%.2f_%.2f_%02i.mat', ...
                              resdir, n, m, k, dens, noise, iter); % Formatstring for saving the results
            
            res = runRankAlgos(algos, B, res, 'unif', iter, iterNoise, filestr);
            
            fprintf('noise = %.2f, iter = %i Gaussian\n', noise, iter);
            load(sprintf('%s/norm_%03i_%03i_%03i_%.2f_%.2f_%02i.mat', ...
                datadir, n, m, k, dens, noise, iter));
            
            res = runRankAlgos(algos, B, res, 'norm', iter, iterNoise, filestr);
        end
    end
    save([resdir '/big_vary_noise_all' result_suffix '.mat'], 'Noise');
    save([resdir '/big_vary_noise_all' result_suffix '.mat'], '-struct', 'res', ...
         '-append');

end % if tests.vary_noise

end % function run_synth_rank_small
    
function res = initRes(algos, rows, cols)
template1 = struct(...
    'rank', nan(rows, cols), ...
    'nuc',  nan(rows, cols), ...
    'LL',   nan(rows, cols), ...
    'time', nan(rows, cols) );
template = struct();
template.norm = template1;
template.unif = template1;

res = struct();
if algos.nuclear
    res.nuclear = template;
end
if algos.SVD
    res.SVD = template;
end
if algos.proje
    res.proje = template;
end
if algos.shay
    res.shay = template;
end
if algos.lpca
    res.lpca = template;
    % LPCA also stores the iterations
    res.lpca.norm.iters = nan(rows, cols);
    res.lpca.unif.iters = nan(rows, cols);
end
if algos.LB
    res.LB = template;
end
end % function initRes

function res = runRankAlgos(algos, B, res, type, row, col, filestr, varargin)
    defaultThreshold  = 0.5;
    defaultReplicates = 5;
    defaultVerbose    = false;
    p = inputParser;
    addRequired(p, 'algos', @(x) ...
                validateattributes(x, {'struct'}, {'nonempty'}, 'runRankAlgos', ...
                                   'algos', 1));
    addRequired(p, 'B', @(x) validateattributes(x, {'numeric'}, {'2d', ...
                        'binary'}, 'runRankAlgos', 'B', 2));
    addRequired(p, 'res', ...
                @(x) validateattributes(x, {'struct'}, {'nonempty'}, ...
                                        'runRankAlgos', 'res', 3));
    addRequired(p, 'type', ...
                @(x) any(validatestring(x, {'norm', 'unif'}, 'runRankAlgos', ...
                                    'type', 4)));
    addRequired(p, 'row', ...
                @(x) validateattributes(x, {'numeric'}, {'scalar', 'integer', ...
                   'positive'}, 'runRankAlgos', 'row', 5));
    addRequired(p, 'col', ...
                @(x) validateattributes(x, {'numeric'}, {'scalar', 'integer', ...
                   'positive'}, 'runRankAlgos', 'col', 6));
    addRequired(p, 'filestr', @ischar);
    addOptional(p, 'threshold', defaultThreshold, ...
                @(x) validateattributes(x, {'numeric'}, {'scalar', 'real'}, ...
                                        'runRankAlgos', 'threshold'));
    addOptional(p, 'replicates', defaultReplicates, ...
                @(x) validateattributes(x, {'numeric'}, {'scalar', 'integer' ...
                   'positive'}, 'runRankAlgos', 'replicates'));
    addOptional(p, 'verbose', defaultVerbose, ...
                @(x) validateattributes(x, {'numeric', 'logical'}, {'scalar'}, ...
                                        'runRankAlgos', 'verbose'));
    parse(p, algos, B, res, type, row, col, filestr, varargin{:});
    
    tau = p.Results.threshold;
    replicates = p.Results.replicates;
    verbose = p.Results.verbose;

    
    if algos.nuclear
        fprintf('\tnuclear\n');
        tic
        [r, X] = RoundingRankNuclear(B, 'verbose', verbose, 'threshold', tau);
        t = toc;
        res.nuclear.(type).rank(row, col) = r;
        res.nuclear.(type).nuc( row, col) = norm(svd(X), 1); % Nuclear norm
        res.nuclear.(type).LL(  row, col) = loglikelihood(B, X); 
        res.nuclear.(type).time(row, col) = t;
        save(sprintf(filestr, 'nuclear', type), 'r', 'X');
    end
    
    if algos.SVD
        fprintf('\tSVD\n');
        tic
        [r, X] = RoundingRankSVD(B, 'verbose', verbose, 'threshold', tau);
        t = toc;
        res.SVD.(type).rank(row, col) = r;
        res.SVD.(type).nuc( row, col) = norm(svd(X), 1); % Nuclear norm
        res.SVD.(type).LL(  row, col) = loglikelihood(B, X); 
        res.SVD.(type).time(row, col) = t;
        save(sprintf(filestr, 'svd', type), 'r', 'X');
    end
    
    if algos.proje
        fprintf('\tprojection\n');
        tic
        [r, X] = RoundingRankProj(B, 'verbose', verbose, 'threshold', ...
                                  tau, 'replicates', replicates, 'solver', ...
                                  'gurobi');
        t = toc;
        res.proje.(type).rank(row, col) = r;
        res.proje.(type).nuc( row, col) = norm(svd(X), 1); % Nuclear norm
        res.proje.(type).LL(  row, col) = loglikelihood(B, X); 
        res.proje.(type).time(row, col) = t;
        save(sprintf(filestr, 'proje', type), 'r', 'X');
    end
    
    if algos.lpca
        fprintf('\tLogistic PCA\n');
        tic
        [r, X, iters] = RoundingRankLPCA(B, 'verbose', verbose, 'threshold', ...
                                         tau, 'replicates', replicates, ...
                                         'numiterations', 50);
        t = toc;
        res.lpca.(type).rank( row, col) = r;
        res.lpca.(type).nuc(  row, col) = norm(svd(X), 1); % Nuclear norm
        res.lpca.(type).LL(   row, col) = loglikelihood(B, X); 
        res.lpca.(type).time( row, col) = t;
        res.lpca.(type).iters(row, col) = iters;
        save(sprintf(filestr, 'lpca', type), 'r', 'X');
    end            
    
    if algos.shay
        fprintf('\tShay''s\n');
        warning('Shay''s cannot handle non-square matrices!');
        tic
        r = Shays(B);
        t = toc;
        disp(r); % DEBUG
        res.shay.(type).rank(row, col) = r;
        %shay.(type).nuc( row, col) = norm(svd(X), 1); % Nuclear norm
        res.shay.(type).time(row, col) = t;
        save(sprintf(filestr, 'shay', type), 'r');
    end
    
    if algos.LB
        fprintf('\tLowerBound\n');
        tic
        r = RRLowerBound(B);
        t = toc;
        res.LB.(type).rank(row, col) = r;
        res.LB.(type).time(row, col) = t;
        save(sprintf(filestr, 'LB', type), 'r');
    end
    
end
