function [bestB, bestX, bestT] = sparse_asso(A, T, K, varargin)
% SPARSE_ASSO - computes BMF for sparse matrices using ASSO
%
% Usage
%  [B, C] = SPARSE_ASSO(A, T, k)
%    Returns rank-k factorization trying all values from T
%  [B, C, t] = SPARSE_ASSO(A, T, k)
%    Returns also the value of t used.
%  [B, C] = SPARSE_ASSO(A, T, k, 'keyword', 'value')
%    assigns value 'value' to keyword 'keyword'. Possible keywords
%    are:
%	KEYWORD			MEANING
%    'VerbosityLevel'       Controls the level of verbosity. 0 is no
%                   prints (default); 1 is some, 2 is more, etc.
%
%    'processes'            Number of processes to launch (default
%                           = 2)
%
% See also: ASSO, PANDA

error(nargchk(3, Inf, nargin));
if mod(size(varargin, 2), 2) ~= 0,
    error('Arguments must be Keyword-Value pairs');
end

% Input parameter checking
if any(A(A ~= 0) ~= 1)
    error('Input matrix A must be 0/1');
end

% Check if A is sparse
if ~issparse(A)
    warning('Input matrix is not sparse, consider using ASSO instead');
    A = sparse(A);
end

%iA = uint8(A);

if ~isvector(T)
    error('Parameter T must be a scalar or vector')
end

if ~isscalar(K)
    error('Parameter k must be a scalar')
end

[n, m] = size(A);

% Defaults
verbosity = 0;
processes = 2;

%% Supported keywords
keywordNames = {'verbosityLevel', 'processes'};


for i=1:2:(size(varargin,2)-1),
    switch lower(make_canonical(varargin{i}, keywordNames))
        case {'verbositylevel'}
            verbosity = varargin{i+1};
        case {'processes'}
            processes = varargin{i+1};
        otherwise
            error(['Unknown parameter: ' varargin{i}]);
    end
end


% Temp check
if processes < 2
    warning('Non-parallel running is not optimized');
end



% Store space for output variables

bestT = -1;
bestB = [];
bestX = [];
bestErr = Inf;


%% To the actual computation

Corig = full(A*A'); % Association confidences
Corig = Corig./repmat(diag(Corig)', n, 1); % Normalize

%% DEBUG
%disp('Corig')
%disp(Corig)
%% END DEBUG

% Iterate over all values in T
for titer=1:length(T)
    t = T(titer);
    if (verbosity > 0),
        fprintf(1, 't = %f', t);
    end
    B = nan(n,K);
    X = nan(K,m);
    available_cols = 1:n;
    C = sparse(Corig >= t);
    %% DEBUG
    %disp(t)
    %disp(C)
    %% END DEBUG
    mask = zeros(n, m, 'uint8');
    for k=1:K,
        if (verbosity > 1),
            fprintf(1, '  k = %i', k);
        end
        %% select column of C (from available columns) to B and
        %% corresponding row to X. avail_id is index to C.
        %% If processes > 1, use parallel version
        %if processes > 1,
        [avail_id, r] = select_best_column_sparse_par(A, ...
            C, ...
            available_cols, ...
            mask, ...
            processes);
        B(:,k) = full(double(C(:,available_cols(avail_id))));
        X(k,:) = r;
        %% DEBUG
        %disp('B')
        %disp(B)
        %disp('X')
        %disp(X)
        %% END DEBUG
        mask = max(mask, uint8(bprod(B(:,k), X(k,:))));
        % Update available_cols;
        foo = [available_cols(1:avail_id-1) available_cols(avail_id+1:end)];
        available_cols = foo;
        % DEBUG
        %disp(length(available_cols))
        % END DEBUG
    end %% for k=1:maxK
    err = bnorm(A, B, X);
    if err < bestErr
        if (verbosity > 2)
            fprintf(1, 'Error improved from %u to %u\n', bestErr, ...
                err);
        end
        bestErr = err;
        bestB = B;
        bestX = X;
        bestT = t;
    end
    if (verbosity > 0),
        fprintf(1, '\n');
    end
end %% for t=T

